/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "global.h"
#include "agent.h"
#include "simnode.h"
#include "link.h"
#include "roadstart.h"
#include "roadend.h"
#include "world.h"
#include "network.h"
#include "clock.h"
#include "trip.h"
#include "serializer.h"
#include "communicator.h"
#include <cassert>
#include <algorithm>


using namespace std;

namespace allsim {
namespace microsim {

unsigned const Agent::ENTER_LINK = 0;
unsigned const Agent::LEAVE_LINK = 1;
unsigned const Agent::START_TRIP = 2;

unsigned Agent::creationCounter = 0;



Agent::Agent(unsigned id, World &theWorld)
:deleted(false),id(id),myWorld(theWorld), myClock(theWorld.getClock()),
currentRoadID(numeric_limits<unsigned>::max()),
nextRoadID(numeric_limits<unsigned>::max()),
currentRoadEnd(NULL),
nextRoadStart(NULL),
freeSpeedLeavingTime(-numeric_limits<double>::max()),
timerTime(-numeric_limits<double>::max()),
nextTimerTime(-numeric_limits<double>::max()),
nextTimerType(numeric_limits<unsigned>::max()),
lastRoadRegisterTime(-numeric_limits<double>::max()),
currentTripIndex(0),
squeezeFlag(false)
{
// 	if(id == AGENTID2)
// 	{
// 		cerr << "agent " << id << " was created on proc " << Communicator::getProcId() << endl;
// 	}
	++creationCounter;
}


Agent::~Agent()
{
// 	if(id == AGENTID2)
// 	{
// 		cerr << "agent " << id << " was deleted on proc " << Communicator::getProcId() << endl;
// 	}
	// deleted is only for debugging purposes
	// using this "trick" we are often able to detect double deletions...
	if(deleted)
	{
		cerr << "error: this agent was already deleted..." << endl;
		throw;
	}
	deleted = true;
	// the Agent is responsible for deleting its trips
	deleteAllTrips();
}

void Agent::handleTimeToEnterRoad(double time)
{
// 	if(id== AGENTID1)
// 	{
// 		cout.precision(DISP_PREC);
// 		cout << "agent " << id << " got time leave link " << currentRoadID << " and enter link " << nextRoadID  
// 		     << " at " << time << endl;
// 	}

	// debug:
	if(myClock.getLastTimerTime() > time)
	{
		cerr << "warning: lastTimerTime > time" << endl
		     << "lastTimerTime=" << myClock.getLastTimerTime()
		     << ", time=" << time << endl
		     << "agentid = " << id << endl;
	}

	nextTimerTime = time;
	nextTimerType = ENTER_LINK;
	myClock.registerTimer(this, nextTimerTime);
}

void Agent::handleEarliestTimeToLeaveRoad(double time)
{
// 	if(id== AGENTID1)
// 	{
// 		cout.precision(DISP_PREC);
// 		cout << "agent " << id << " got earliest time to leave link "
// 		     << nextRoadID << " at " << time << endl;
// 	}
	// compare the earliest time to leave the link 
	// with constraints due to travel speeds and register a timer accordingly
	double earliestAgentLeavingTime = max(time, freeSpeedLeavingTime);
	nextTimerTime = earliestAgentLeavingTime;
	nextTimerType = LEAVE_LINK;
	myClock.registerTimer(this, nextTimerTime);
}

unsigned Agent::getFirstLinkId() const
{
	if(!trips.empty())
		return	trips[0]->getStartLinkId();
	else
		return numeric_limits<unsigned>::max();
}

double Agent::getLastRegisterTime() const
{
	return lastRoadRegisterTime;
}

char Agent::getActivityType(unsigned index) const
{
	if(index < actTypes.size())
		return actTypes[index];
	else
		return '?';
}

Trip const *Agent::getTrip(unsigned index) const
{
	if(index < trips.size())
		return trips[index];
	else
		return NULL;
}

void Agent::handleTimeToStartTrip(double time)
{
	nextTimerTime = time;
	nextTimerType = START_TRIP;
	myClock.registerTimer(this, nextTimerTime);
}


void Agent::timerExpired()
{
	unsigned timerType = nextTimerType;	
	timerTime = nextTimerTime;

	if(DEBUG && !ENTRY_ONLY && (ALL || nextRoadID == LINKID1 || nextRoadID == LINKID2 || id == AGENTID1 || id == AGENTID2))
		if(timerType == LEAVE_LINK)
		{
			cout.precision(DISP_PREC);
			cout << "z a" << id << " r" 
			     << currentRoadID << " t" << timerTime << endl;
		}

	// there is no matsim event when an agent reaches the end of a road
	//	if(WRITE_MATSIM_EVENTS)


	
	switch(timerType)
	{
		case ENTER_LINK:
			// this means: we are ready to enter the next link
			// first notify old link that we are leaving
			
			if(currentRoadEnd)
				currentRoadEnd->noteFirstAgentLeavesRoad(this, timerTime);
			// compute earliest arrival time at end of current link due to free speed
			freeSpeedLeavingTime = myWorld.getRoadStart(nextRoadID)->getFreeSpeedLeavingTime(timerTime);
//			freeSpeedLeavingTime = nextRoadStart->getFreeSpeedLeavingTime(timerTime);

			// check if the trip is done
			if(trips[currentTripIndex]->isDone())
			{
				// inform the road that we are not actually entering but ending our trip here...
				if(nextRoadStart)
					nextRoadStart->noteAgentEndsTrip(this, timerTime);
			}
			else
			{
				// inform the new link that we are entering...
				if(nextRoadStart)
					nextRoadStart->noteAgentEntersRoad(this, timerTime);
			}


			//give the links the chance to react to the new situation
			if(nextRoadStart)
				nextRoadStart->checkTasks();
			if(currentRoadEnd)
				currentRoadEnd->checkTasks();
	
			// make the new link the current link
			currentRoadID = nextRoadID;
			// set nextRoadID to a dummy value since we don't know the next ID yet
			nextRoadID = numeric_limits<unsigned>::max();

			// the purpose of the following line is to produce a segmentation fault if 
			// anybody tries to acess the current Road End. Remove this line if
			// you are sure that your code is correct. David Charypar, 2007-01-09
			currentRoadEnd = reinterpret_cast<RoadEnd*>(-1);
			// currentRoadEnd = myWorld.getRoadEnd(currentRoadID);

			if(trips[currentTripIndex]->isDone())
				processNextTrip();

		break;
		case LEAVE_LINK:
			// this means: we are ready to leave the link
			// but first we need to find out where we are and where to go
			currentRoadEnd = myWorld.getRoadEnd(currentRoadID);
			computeNextRoad();

			// at this point we know the new link so we can register to it...
			if(nextRoadStart)
			{
				lastRoadRegisterTime = timerTime;
				nextRoadStart->registerAgentForEntering(this, timerTime);
			}
			else
			{
				// this should not happen as trip ends are handled in case ENTER_LINK:
				throw;
			}
			//give the links the chance to react to the new situation
			if(nextRoadStart)
				nextRoadStart->checkTasks();
			if(currentRoadEnd)
				currentRoadEnd->checkTasks();

		break;
		case START_TRIP:
			// this means: we are ready to start the next trip
			// but first we need to compute where to go
			computeNextRoad();

			// at this point we know the new link so we can register to it...
			if(nextRoadStart)
			{
				// register this agent with the first road of this trip
				lastRoadRegisterTime = timerTime;
				nextRoadStart->registerAgentForStartingATrip(this, timerTime);
			}

			//give the link the chance to react to the new situation
			if(nextRoadStart)
				nextRoadStart->checkTasks();

			// if there was no next link , i.e. the current trip is done
			// --> look for oncoming trips and take care of them
			if(!nextRoadStart)
				processNextTrip();
	}
}

int Agent::getSortPriority() const
{
	return -static_cast<int>(id);
}

bool Agent::getSqueezeFlag() const
{
	return squeezeFlag;
}

void Agent::setSqueezeFlag(bool newState)
{
	squeezeFlag = newState;
}

unsigned Agent::getCurrentTripIndex()
{
	return currentTripIndex;
}


unsigned Agent::getId() const
{
	return id;
}

void Agent::deleteAllTrips()
{
	while(!trips.empty())
	{
		Trip *deletableTrip;
		deletableTrip = trips.back();
		trips.pop_back();
		delete deletableTrip;
	}
}


void Agent::addTrip(Trip *trip)
{
	trips.push_back(trip);
}

void Agent::addActivityType(char actType)
{
	actTypes.push_back(actType);
}


void Agent::deduceEndLinksOfTrips()
{
	if(trips.size() == 0)
		return;
	// the trips currently lack their links.
	// copy the start link of the next trip to each trip, respectively
	for(unsigned i = 0; i < trips.size()-1; ++i)
	{
		assert(trips[i+1]->getSize()>0);
		trips[i]->addLinkIdToBackOfRoute(trips[i+1]->getStartLinkId());
	}
	// dont forget the end link of the last trip
	assert(trips[0]->getSize()>1);
	trips[trips.size()-1]->addLinkIdToBackOfRoute(trips[0]->getStartLinkId());
}


void Agent::startActing()
{
	// check the first trip and register a timer for the departure time
	if(currentTripIndex < trips.size())
		handleTimeToStartTrip(trips[currentTripIndex]->getDepartureTime());
}

void Agent::serializeToVector(vector<char> *theVector) const
{
//	cerr << "(" << Communicator::getProcId() << "): serializing Agent with id " << id << endl;
	Serializer::setVector(theVector);

	Serializer::serializeToVector(id);
	Serializer::serializeToVector(deleted);
	Serializer::serializeToVector(currentRoadID);
	Serializer::serializeToVector(nextRoadID);
	Serializer::serializeToVector(freeSpeedLeavingTime);
	Serializer::serializeToVector(timerTime);
	Serializer::serializeToVector(nextTimerTime);
	Serializer::serializeToVector(nextTimerType);
	Serializer::serializeToVector(lastRoadRegisterTime);
	Serializer::serializeToVector(static_cast<unsigned>(trips.size()));
	for(vector<Trip*>::const_iterator ti = trips.begin();
	                             ti != trips.end(); ++ti)
		(*ti)->serializeToVector(theVector);
	
	Serializer::serializeToVector(currentTripIndex);
	Serializer::serializeToVector(squeezeFlag);
}

vector<char>::const_iterator Agent::unserializeFromVectorIterator(vector<char>::const_iterator iter)
{
	unsigned tripsSize;

	iter = Serializer::unserializeFromVectorIterator(deleted, iter);
	iter = Serializer::unserializeFromVectorIterator(currentRoadID, iter);
	iter = Serializer::unserializeFromVectorIterator(nextRoadID, iter);
	iter = Serializer::unserializeFromVectorIterator(freeSpeedLeavingTime, iter);
	iter = Serializer::unserializeFromVectorIterator(timerTime, iter);
	iter = Serializer::unserializeFromVectorIterator(nextTimerTime, iter);
	iter = Serializer::unserializeFromVectorIterator(nextTimerType, iter);
	iter = Serializer::unserializeFromVectorIterator(lastRoadRegisterTime, iter);
	iter = Serializer::unserializeFromVectorIterator(tripsSize, iter);

	// make sure there are not trips left
	assert(trips.size() == 0);

	for(unsigned i = 0; i < tripsSize; ++i)
	{
		Trip *unserializedTrip;
		iter = Trip::unserializeFromVectorIterator(unserializedTrip, iter);
		addTrip(unserializedTrip);	
	}

 	iter = Serializer::unserializeFromVectorIterator(currentTripIndex, iter);
 	iter = Serializer::unserializeFromVectorIterator(squeezeFlag, iter);

	// create the neede pointers to last road end and next road start
	if(currentRoadID == numeric_limits<unsigned>::max())
		currentRoadEnd = NULL;
	else
		currentRoadEnd = myWorld.getRoadEnd(currentRoadID);
	
	if(nextRoadID == numeric_limits<unsigned>::max())
		nextRoadStart = NULL;
	else
		nextRoadStart = myWorld.getRoadStart(nextRoadID);	

	return iter;
}

unsigned Agent::getNumTrips() const
{
	return trips.size();
}

unsigned Agent::getTotalLengthOfRoutes() const
{
	unsigned counter = 0;
	for(unsigned i = 0; i < trips.size(); ++i)
		counter += trips[i]->getSize();
	return counter;
}


unsigned Agent::getNumCreatedAgents()
{
	return creationCounter;
}

bool Agent::isDeleted()
{
	return deleted;
}

void Agent::computeNextRoad()
{
	if(currentTripIndex < trips.size())
	{
		nextRoadID = trips[currentTripIndex]->getNextLinkIdAndAdvance();
		nextRoadStart = myWorld.getRoadStart(nextRoadID);
//		cerr << "agid"  << id << ": nextRoadID " << nextRoadID << ", nextRoadStart " << nextRoadStart << " id " << nextRoadStart->getId() << endl;
	}
}

void Agent::processNextTrip()
{
	if(currentTripIndex < trips.size()-1)
	{
		// there is a next trip
		++currentTripIndex;
		handleTimeToStartTrip(max(trips[currentTripIndex]->getDepartureTime(), timerTime));
		currentRoadID = numeric_limits<unsigned>::max();
		currentRoadEnd = NULL;
	}
	else
	{
		// there is no next trip and this agent is done. Inform world about it
		currentRoadID = numeric_limits<unsigned>::max();
		currentRoadEnd = NULL;
		myWorld.noticeAgentFinished(id);
	}
}





}
}
