/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "clock.h"
#include "timer.h"
#include "callable.h"
#include "world.h"
#include "serializer.h"
#include "communicator.h"
#include "microsimconfig.h"
#include "toolsglobal.h"
#include <cassert>
#include <cmath>
#include <ctime>
#include <iostream>
#include <set>

using namespace std;

namespace allsim {
namespace microsim {

Clock::Clock(World &theWorld)
:maxTime(-numeric_limits<double>::max()), 
numProcessedTimers(0), 
myWorld(theWorld)
{
}

Clock::~Clock()
{
}


void Clock::registerTimer(Callable *callable, double time)
{
	if(callable == NULL)
		throw "error in Clock::registerTimer(): callable == NULL!";
	if(!(time >= maxTime))
	{
		cerr << "error in register Timer: trying to register timer in the past!" << endl;
		cerr << "time = " << time << ", maxTime = " << maxTime << endl;
		throw;
	}
	timerQueue.push(Timer(callable, time));
}

void Clock::processTimers()
{
	ostream &logFile = Communicator::getLogFile();
	cerr.precision(DISP_PREC);
	cout.precision(DISP_PREC);
	using namespace config;
	using namespace tools;
	double const SIM_END_TIME = convertTimeStringToTime(Config::getModuleConfig(Config::DEQSIM).getParam(Config::DEQSIM_ENDTIME));

	if(DEBUG)
	{
		logFile << "inside void Clock::processTimers()" << endl;
		logFile << "timerQueue.size() is " << timerQueue.size() << endl;
	}
	double lastDisplayedSimulationTime = -1e100;

	time_t startTime = time(NULL);
	time_t lastDisplayTime = startTime;
	time_t currentTime = startTime;
	double deltaDisplay = 1.0;
	double currentDelta;

	double lowestRTR = numeric_limits<double>::infinity();

	while(!timerQueue.empty())
	{
		Timer const &currentTimer = timerQueue.top();
		maxTime = currentTimer.getTime();
		Callable *currentCallable = currentTimer.getCallable();
		// remove timer from data structures and delete it...
		timerQueue.pop();
		
		if(maxTime > SIM_END_TIME)
 			break;
		currentCallable->timerExpired();
		

 		if(numProcessedTimers % 8192 == 0)
		{
			currentTime = time(NULL);
			if((currentDelta = difftime(currentTime, lastDisplayTime))>= deltaDisplay)			
			{
				double realTimeRatio = (maxTime - lastDisplayedSimulationTime) / currentDelta;
				lowestRTR = min(lowestRTR, realTimeRatio);
				
				logFile << "P" << Communicator::getProcId();
				logFile << ": tsim=" << maxTime/3600.0 << "h,     \tRTR=" << realTimeRatio;
				logFile << "       \tProcTime=";
				logFile << difftime(currentTime, startTime) << "s.";
				
				logFile << " nAgents=" << myWorld.countAgents();
				logFile << " nPendingTimers=" << timerQueue.size(); 
				logFile << " nAgentDone=" << myWorld.getNumFinishedAgents() << endl;
				lastDisplayedSimulationTime = maxTime;
				lastDisplayTime = currentTime;
			}
		}
		++numProcessedTimers;
	}

	currentTime = time(NULL);

	logFile << "P" << Communicator::getProcId();
	logFile << ": tsim=" << maxTime/3600.0 << "h,";
	logFile << "       \tProcTime=";
	logFile << difftime(currentTime, startTime) << "s.";

	logFile << " nAgents=" << myWorld.countAgents();
	logFile << " nPendingTimers=" << timerQueue.size();
	logFile << " nAgentDone=" << myWorld.getNumFinishedAgents() << endl;


	logFile << "processTimers() ran ";
	logFile << difftime(currentTime, startTime);
	logFile << " seconds" << endl;
	logFile << "lowest real time ratio was " << lowestRTR << endl;
}

unsigned int Clock::getNumProcessedTimers()
{
	return numProcessedTimers;
}

unsigned int Clock::getNumTimersToDo()
{
	return timerQueue.size();
}

double Clock::getLastTimerTime()
{
	return maxTime;
}

}

} // end namespace allsim
