/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "communicator.h"
#include "microsimconfig.h"
#include "global.h"
#include <iostream>
#include <sstream>

#ifdef PARALLEL
#include <mpi.h>
#endif


using namespace std;

namespace allsim {
namespace microsim {


using namespace config;

int Communicator::numProcs, Communicator::myProcId;
char *buffer = NULL;
unsigned const BUFFSIZE = 1024U*1024U;
ofstream Communicator::eventsFile;
ofstream Communicator::comLogFile;
ofstream Communicator::logFile;
bool Communicator::eventsFileIsBinary;
int Communicator::byteOrder64[8];
int Communicator::byteOrder32[4];
int Communicator::byteOrder16[2];

#ifdef PARALLEL
// this is parallel mpi code
void Communicator::init(int &nArgs, char **&args)
{
	if(DEBUG)
		cerr << "inside void Communicator::init()" << endl;

	initCommon();

	// call MPI::Init passing command line parameters.
	MPI::Init(nArgs, args);
	numProcs = MPI::COMM_WORLD.Get_size();
	myProcId = MPI::COMM_WORLD.Get_rank();

	stringstream logFileName;
	logFileName << "deqsim.log." << myProcId;
	logFile.open(logFileName.str().c_str());
	logFile.precision(10);

	if(WRITE_DUMMY_TVEH)
	{
		stringstream snapshotFilename;
		snapshotFilename << config::Config::getModuleConfig(Config::SCENARIO)
		.getParam(Config::SCENARIO_WORIKINGSNAPSHOTFILENAME);
		snapshotFilename << "." << myProcId;
		logFile << "on proc " << myProcId << ": creating dummy snapshotfile " << snapshotFilename.str() << endl;
		ofstream dummy(snapshotFilename.str().c_str());
		dummy.close();
	}

	// open events file
	if(WRITE_MATSIM_EVENTS)
	{
		stringstream eventsFilename;
		eventsFilename << config::Config::getModuleConfig(Config::EVENTS).getParam(Config::EVENTS_OUTPUTFILE);
		eventsFilename << "." << myProcId;
		if(Config::getModuleConfig(Config::EVENTS).getParam(Config::EVENTS_OUTPUTFORMAT).compare("matsimDEQ1") == 0)
		{
			eventsFileIsBinary = true;
			eventsFile.open(eventsFilename.str().c_str(), ofstream::binary | ofstream::out | ofstream::trunc);
		}
		else
		{
			eventsFileIsBinary = false;
			eventsFile.open(eventsFilename.str().c_str());
		}
	}

	if(LOG_COMM)
	{
		stringstream comLogFileName;
		comLogFileName << "comlog.txt." << myProcId;
		comLogFile.open(comLogFileName.str().c_str());
		comLogFile.precision(20);
	}

	// allocate some buffer space for buffered communications
//	unsigned const BUFFSIZE=1024U*1024U;
//	buffer = new char[BUFFSIZE];
//	MPI::Attach_buffer(buffer, BUFFSIZE); 

	if(DEBUG)
	{
		if(numProcs == 1)
		{
			if(myProcId == 0)
				cerr << "There is " << numProcs << " process running..." << endl;
			logFile << "There is " << numProcs << " process running..." << endl;
			logFile << "my process id is " << myProcId << "." << endl;
		}
		else
		{
			if(myProcId == 0)
				cerr << "There are " << numProcs << " processes running..." << endl;
			logFile << "There are " << numProcs << " processes running..." << endl;
			logFile << "my process id is " << myProcId << "." << endl;
		}
	}
}

void Communicator::close()
{
	// close events file
	eventsFile.close();
	comLogFile.close();
	logFile.close();

// 	void *dummy;
// 	MPI::Detach_buffer(dummy); 
// 	delete buffer;
// 	buffer = NULL;
	MPI::Finalize();
}



void Communicator::sendToAllNeighbors(vector<char> &buffer, int messageTag)
{
	// skip this step in single process mode
	if(numProcs <= 1)
		return;

	// since we are doing a nonblocking send, the sendbuffer has to be static...
	static char *sendBuf = NULL;
	unsigned sendCounter = 0;
	
	// prepare the send buffer for MPI;
	if(sendBuf != NULL)
		delete sendBuf;
	sendBuf = new char[buffer.size()];
	copy(buffer.begin(), buffer.end(), sendBuf);
	
	for(int curProc = 0; curProc < numProcs; ++curProc)
	{
		if(curProc != myProcId)
		{
//			cerr << "(" << Communicator::getProcId() << "): ";
//			cerr << "sendig message to proc " << curProc << endl;
			MPI::COMM_WORLD.Isend(sendBuf, buffer.size(), MPI::CHAR, curProc, messageTag);
			sendCounter += buffer.size();
		}
	}

}

void Communicator::receiveFromAllNeighbors(vector<char> &buffer, int messageTag)
{
	// skip this step in single process mode
	if(numProcs <= 1)
		return;
	// here we receive a message from all our neighbors
	
	char *charbuf = NULL;
	MPI::Status messageStatus;
	unsigned receiveCounter = 0;

	unsigned nNeighbors = numProcs-1;
	for(unsigned i = 0; i < nNeighbors; ++i)
	{
//		cerr << "(" << Communicator::getProcId() << "): ";
//		cerr << "waiting for message # " << i << endl;
		
		// wait for the next message and get its source, tag and size
		MPI::COMM_WORLD.Probe(MPI::ANY_SOURCE, messageTag, messageStatus);
		// allocate memory for the message...
		charbuf = new char[messageStatus.Get_count(MPI::CHAR)];
		// actually receive the message
//		cerr << "(" << Communicator::getProcId() << "): ";
//		cerr << "receiving message from " << messageStatus.Get_source();
//		cerr << " with size " << messageStatus.Get_count(MPI::CHAR) << endl;
		MPI::COMM_WORLD.Recv(charbuf, messageStatus.Get_count(MPI::CHAR), MPI::CHAR, messageStatus.Get_source(), messageStatus.Get_tag());

		receiveCounter += messageStatus.Get_count(MPI::CHAR);

		// copy the message to a vector ...
		unsigned oldSize = buffer.size();
		buffer.resize(oldSize + messageStatus.Get_count(MPI::CHAR));
		copy(charbuf, charbuf+messageStatus.Get_count(MPI::CHAR), buffer.begin()+oldSize);

		// free the receive buffer
		delete charbuf;
	}
	// block until all communication of agents is finished...
	MPI::COMM_WORLD.Barrier();
//	cerr << "(" << Communicator::getProcId() << "): ";
//	cerr << "Communicator::receiveFromAllNeighbors() received " << buffer.size() << endl; 
}	

void Communicator::sendToProcess(int procId, int messageTag, vector<char> &message, double time)
{
	// this is a blocking send....
	char *sendBuf = new char[message.size()];
	copy(message.begin(), message.end(), sendBuf);
	if(LOG_COMM)
	{
		if(messageTag % 2 == 0)
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " rsP" << myProcId << "->reP" << procId 
				<< " size:" << message.size() << endl;
		else	
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " reP" << myProcId << "->rsP" << procId 
				<< " size:" << message.size() << endl;
	}
//	MPI::COMM_WORLD.Bsend(sendBuf, message.size(), MPI::CHAR, procId, messageTag);
	MPI::COMM_WORLD.Send(sendBuf, message.size(), MPI::CHAR, procId, messageTag);
	if(LOG_COMM)
	{
		if(messageTag % 2 == 0)
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " rsP" << myProcId << "->reP" << procId 
				<< " size:" << message.size() << " done." << endl;
		else	
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " reP" << myProcId << "->rsP" << procId 
				<< " size:" << message.size() << " done." << endl;
	}
	delete sendBuf;
}

void Communicator::receiveFromProcess(int procId, int messageTag, vector<char> &message, double time)
{
	// this is a blocking receive....
	MPI::Status messageStatus;
	int messageSize;
	if(LOG_COMM)
	{
		if(messageTag % 2 == 1)
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " rsP" << myProcId << "<-reP" << procId 
				<< " size:?" << endl;
		else	
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " reP" << myProcId << "<-rsP" << procId 
				<< " size:?" << endl;
	}
	MPI::COMM_WORLD.Probe(procId, messageTag, messageStatus);
	messageSize = messageStatus.Get_count(MPI::CHAR);
	char *recvbuf = new char[messageSize];
	if(LOG_COMM)
	{
		if(messageTag % 2 == 1)
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " rsP" << myProcId << "<-reP" << procId 
				<< " size:?" << endl;
		else	
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " reP" << myProcId << "<-rsP" << procId 
				<< " size:" << messageSize << endl;
	}
	MPI::COMM_WORLD.Recv(recvbuf, messageSize, MPI::CHAR, procId, messageTag);

	if(LOG_COMM)
	{
		if(messageTag % 2 == 1)
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " rsP" << myProcId << "<-reP" << procId 
				<< " size:" << messageSize << " done." << endl;
		else	
			comLogFile << time << ": lID:" << messageTag / 2 
				<< " reP" << myProcId << "<-rsP" << procId 
				<< " size:" << messageSize << " done." << endl;
	}
	message.resize(messageSize);
	copy(recvbuf, recvbuf+messageSize, message.begin());

	delete recvbuf;
}

void Communicator::waitForOtherProcesses()
{
	MPI::COMM_WORLD.Barrier();
}


int Communicator::getMaxmumTag()
{
	return MPI::TAG_UB;
}

#endif

#ifndef PARALLEL
// this is the single cpu code not using mpi
void Communicator::init(int &/*nArgs*/, char **&/*args*/)
{
	initCommon();

	numProcs = 1;
	myProcId = 0;

	logFile.open("deqsim.log");
	logFile.precision(10);

	// create a dummy snapshot file
	if(WRITE_DUMMY_TVEH)
	{
		ofstream dummy(config::Config::getModuleConfig(Config::SCENARIO).
		               getParam(Config::SCENARIO_WORIKINGSNAPSHOTFILENAME).c_str());
		dummy.close();
	}

	// open events file
	if(WRITE_MATSIM_EVENTS)
	{
		if(Config::getModuleConfig(Config::EVENTS).getParam(Config::EVENTS_OUTPUTFORMAT).compare("matsimDEQ1") == 0)
		{
			eventsFileIsBinary = true;
			eventsFile.open(config::Config::getModuleConfig(Config::EVENTS).
			                getParam(Config::EVENTS_OUTPUTFILE).c_str(), ofstream::binary |
					ofstream::out | ofstream::trunc);
		}
		else
		{
			eventsFileIsBinary = false;
			eventsFile.open(config::Config::getModuleConfig(Config::EVENTS).
		        	        getParam(Config::EVENTS_OUTPUTFILE).c_str());
		}

	}
}

void Communicator::close()
{
	// close files
	eventsFile.close();
	logFile.close();
}



void Communicator::sendToAllNeighbors(vector<char> &buffer, int messageTag)
{
	// this method should not be used.
	logFile << "error: trying to send to all neighbors in a single cpu version." << endl;
	throw;
}

void Communicator::receiveFromAllNeighbors(vector<char> &buffer, int messageTag)
{
	// this method should not be used.
	logFile << "error: trying to receive from all neighbors in a single cpu version." << endl;
	throw;
}	

void Communicator::sendToProcess(int procId, int messageTag, vector<char> &message, double time)
{
	// this method should not be used.
	logFile << "error: trying to send to other process in single cpu version." << endl;
	throw;
}

void Communicator::receiveFromProcess(int procId, int messageTag, vector<char> &message, double time)
{
	// this method should not be used.
	logFile << "error: trying to receive from other process in single cpu version." << endl;
	throw;
}

void Communicator::waitForOtherProcesses()
{
}

int Communicator::getMaxmumTag()
{
	return numeric_limits<int>::max();
}

#endif


void Communicator::initCommon()
{
	// find out endiannes of this system

	int64_t testValue64 = 0x00010203L;
	testValue64 <<= 32;
	testValue64 += 0x04050607L;

	int32_t testValue32 = 0x00010203L;
	int16_t testValue16 = 0x0001;
	
	char *vals = reinterpret_cast<char *>(&testValue64);
	for(int i = 0; i < 8; ++i)
		byteOrder64[i] = static_cast<int>(vals[i]);

	vals = reinterpret_cast<char *>(&testValue32);
	for(int i = 0; i < 4; ++i)
		byteOrder32[i] = static_cast<int>(vals[i]);

	vals = reinterpret_cast<char *>(&testValue16);
	for(int i = 0; i < 4; ++i)
		byteOrder16[i] = static_cast<int>(vals[i]);


	if(DEBUG)
	{
		cerr << "64bit byteorder =";
		for(int i = 0; i < 8; ++i)
			cerr << " " << byteOrder64[i];
		cerr << endl;

		cerr << "32bit byteorder =";
		for(int i = 0; i < 4; ++i)
			cerr << " " << byteOrder32[i];
		cerr << endl;

		cerr << "16bit byteorder =";
		for(int i = 0; i < 2; ++i)
			cerr << " " << byteOrder16[i];
		cerr << endl;
	}

}

int Communicator::getNumProcs()
{
	return numProcs;
}

int Communicator::getProcId()
{
	return myProcId;
}


bool Communicator::isEventsFileBinary()
{
	return 	eventsFileIsBinary;
}

ofstream &Communicator::getEventsFile()
{
	return eventsFile;
}

ofstream &Communicator::getComLogFile()
{
	if(comLogFile)
		return comLogFile;
	else
		throw "in Communicator::getComLogFile(): error, comm log file not created, but requested.";
}

ofstream &Communicator::getLogFile()
{
	if(logFile)
		return logFile;
	else
		throw "in Communicator::getLogFile(): error, log file bad.";
}

void Communicator::write32BitIntegerToStream(int32_t data, ostream &theStream)
{
	char convertedData[4];
	char *origData = reinterpret_cast<char *>(&data);
	for(int i = 0; i < 4; ++i)
		convertedData[byteOrder32[i]] = origData[i];
	theStream.write(reinterpret_cast<char *>(convertedData), 4);
}

void Communicator::writeDoubleToStream(double data, ostream &theStream)
{
	char convertedData[8];
	char *origData = reinterpret_cast<char *>(&data);
	for(int i = 0; i < 8; ++i)
		convertedData[byteOrder64[i]] = origData[i];
	theStream.write(reinterpret_cast<char *>(convertedData), 8);
}

int32_t Communicator::read32BitIntegerFromStream(istream &theStream)
{
	int32_t readValue;
	char buffer[4];
	char *destData = reinterpret_cast<char *>(&readValue);
	theStream.read(buffer, 4);

	for(int i = 0; i < 4; ++i)
	{
		destData[i] = buffer[byteOrder32[i]];
//		if(DEBUG)
//			cerr << static_cast<unsigned>(static_cast<unsigned char>(buffer[i])) << " ";
	}
//	if(DEBUG)
//		cerr << endl;
	
	return readValue;
}

double Communicator::readDoubleFromStream(istream &theStream)
{
	double readValue;
	char buffer[8];
	char *destData = reinterpret_cast<char *>(&readValue);
	theStream.read(buffer, 8);

	for(int i = 0; i < 8; ++i)
	{
		destData[i] = buffer[byteOrder64[i]];
//		if(DEBUG)
//			cerr << static_cast<unsigned>(static_cast<unsigned char>(buffer[i])) << " ";
	}
//	if(DEBUG)
//		cerr << endl;
	
	return readValue;
}


}
}
