/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "world.h"
#include "network.h"
#include "clock.h"
#include "simnode.h"
#include "link.h"
#include "roadstart.h"
#include "roadend.h"
#include "agent.h"
#include "serializer.h"
#include "communicator.h"
#include <cassert>
#include <sstream>
#include <iostream>
#include <cmath>

namespace allsim {
namespace microsim {

using namespace std;

RoadEnd::RoadEnd(unsigned id, unsigned fromNodeId, unsigned toNodeId, 
                 double freeSpeedTravelTime, 
                 double gapTravelTime, double minimalGap, 
                 unsigned sizeOfRoad, World &theWorld)
:id(id),
fromNodeId(fromNodeId),
toNodeId(toNodeId),
sizeOfRoad(max(sizeOfRoad, 1u)),
gapTravelTime(gapTravelTime),
syncTime(min(freeSpeedTravelTime, gapTravelTime)),
minimalGap(minimalGap),
myWorld(theWorld),
myNetwork(*theWorld.getNetwork()),
myClock(theWorld.getClock()),
firstCarNotified(false),
lastLeavingTime(-numeric_limits<double>::max()),
roadQueue(new list<Agent*>),
gapQueue(new list<double>),
nextSyncTime(theWorld.getFirstSyncTime()),
//nextSyncTime(numeric_limits<double>::max()),
isInSync(false),
directyConnected(false)
{
// 	Communicator::getLogFile() << "re: "
// 	<< "id=" << id << " fstt=" << freeSpeedTravelTime << " gtt=" << gapTravelTime 
// 	<< " cap=" << 1/minimalGap << " mg=" << minimalGap << " size=" << sizeOfRoad 
// 	<< " sqt=? synct=" << syncTime << endl;
}


RoadEnd::~RoadEnd()
{
	// only delete the road an gap queue if we are not directly connected to a roadStart
	// otherwise the roadStart object will take care of the deletion of the vectors...
	if(!directyConnected)
	{
		delete roadQueue;
		delete gapQueue;
	}
}

unsigned RoadEnd::getId()
{
	return id;
}

void RoadEnd::noteFirstAgentLeavesRoad(Agent *agent, double time)
{
	// Count this agent on the outgoing link
	if(COMPUTE_LOADS)
	{
		SimNode *toNode = dynamic_cast<SimNode *>(myNetwork.getNodeWithId(toNodeId));
		if(toNode)
			toNode->increaseInVolume();
		else
			throw "Cast error in RoadEnd::noteFirstAgentLeavesRoad()";
	}

//	if((id == LINKID1)||(id == LINKID2))
//	if(agent->getId() == AGENTID1)
//  	{
//  		cout << "agent " << agent->getId() << " leaves link " 
//  		     << id << " at " << time 
//  		     << " from/to nodes: " << Network::getLinkWithId(id)->getFromNode()->getId()
//  		     << "/" << Network::getLinkWithId(id)->getToNode()->getId() << endl;
//  	}
	if(DEBUG && !ENTRY_ONLY && (ALL || id == LINKID1 || id == LINKID2 || agent->getId() == AGENTID1 || agent->getId() == AGENTID2))
	{
		cout.precision(DISP_PREC);
		cout << "l a" << agent->getId() << " r" 
		     << id << " t" << time << endl;
	}

	if(WRITE_MATSIM_EVENTS)
	{
		if(Communicator::isEventsFileBinary())
		{
			Communicator::writeDoubleToStream(time, Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(agent->getId(), Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(id, Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(2, Communicator::getEventsFile()); // leave link
		}
		else
		{
			Communicator::getEventsFile() <<  static_cast<int>(floor(time+0.5)) << "\t" 
		                                      << agent->getId() << "\t" 
		                                      << agent->getCurrentTripIndex() << "\t" 
		                                      << id << "\t"
		                                      << myNetwork.getLinkWithId(id)->getFromNode()->getId() 
		                                      << "\t" << "2\tleave\n";
		}
	}

	if(roadQueue->front() == agent)
	{	
		// store the time of the this departure
		lastLeavingTime = max(lastLeavingTime, time);

		// remove the car from the road queue
		roadQueue->pop_front();
		// the car that leaves the queue leaves a gap that travels backward with a certain predictable speed.
		// We hold a queue of gaps that travel backwards
		// add the new gap to the queue if the car was not squeezed in
		if(agent->getSqueezeFlag() == false)
			gapQueue->push_back(lastLeavingTime + gapTravelTime);
		// We have a new first car on the road that needs to bee notified
		firstCarNotified = false;
	}
	else
	{
		ostringstream errorString;
		errorString << "Error: Agent " << agent->getId() << 
		               " tried to leave road " << id << " without being the first in the row.";
		cerr << errorString.str() << endl;
		throw(errorString.str());
	}
	
}

void RoadEnd::checkTasks()
{
	//	we have to notify the next car on the road so that it can get active...
	notifyFirstCarAboutDepartureTime();
	// if we are directly connecte to partner road start inform it that there might have somthing ganged
	if(directyConnected)
		partnerRoadStart->checkTasksFromPartner();
}

void RoadEnd::checkTasksFromPartner()
{
	//	we have to notify the next car on the road so that it can get active...
	notifyFirstCarAboutDepartureTime();
}


void RoadEnd::timerExpired()
{
	// check if we are directly connected to a road start. If so, don't do anything
	if(!directyConnected)
	{
		// exchange data with road end and register a new timer with the clock
		exchangeDataWithRoadStart();
		nextSyncTime += syncTime;
		myClock.registerTimer(this, nextSyncTime);
		checkTasks();

// 		if(isInSync)
// 		{
// 			exchangeDataWithRoadStartReceive();
// 			nextSyncTime += syncTime;
// 
// 			myClock.registerTimer(this, nextSyncTime - SYNC_OFFSET);
// 			myClock.registerTimer(this, nextSyncTime);
// 			isInSync = false;
// 			checkTasks();
// 		}
// 		else
// 		{
// 			isInSync = true;
// 			exchangeDataWithRoadStartSend();
// 		}
	}
	else
	{
		// we should not get a sync timer if we are directly connected!
		throw;
	}
}

int RoadEnd::getSortPriority() const
{
	return 2 * static_cast<int>(id) + 1;
}

void RoadEnd::registerInitialTimerAtClock()
{
	if(!directyConnected)
	{
		nextSyncTime += syncTime;
//		myClock.registerTimer(this, nextSyncTime - SYNC_OFFSET);
		myClock.registerTimer(this, nextSyncTime);
	}
}

void RoadEnd::exchangeDataWithRoadStart()
{
	// RoadEnds first receive and then send a message
	// receive information about cars that entered road start
	receiveExchangeMessage();
	readLongExchangeMessage();
	exchangeMessageReceive.clear();
	
	// send information about gaps that are travelling backwards through the link
	createLongExchangeMessage();
	emptyGapQueue();
	sendExchangeMessage();
	exchangeMessageSend.clear();
}

void RoadEnd::exchangeDataWithRoadStartSend()
{
	// this method is no longer used
	abort();

	createLongExchangeMessage();
	emptyGapQueue();
	sendExchangeMessage();
}

void RoadEnd::exchangeDataWithRoadStartReceive()
{
	// this method is no longer used
	abort();

	receiveExchangeMessage();
	readLongExchangeMessage();
	exchangeMessageSend.clear();
	exchangeMessageReceive.clear();
}

bool RoadEnd::isDirectlyConnectable()
{
	return myWorld.hasRoadStart(id);
}

bool RoadEnd::isDirectlyConnected()
{
	return directyConnected;
}

void RoadEnd::directlyConnectToRoadStart()
{
	// check if roadStart is also on this process:
	RoadStart *theRoadStart;
	if((theRoadStart = myWorld.getRoadStart(id)) == NULL)
	{
		cerr << "error: roadStart with id " << id << " is not available on this process" << endl;
		throw;
	}
	// delete my own road and qap queue.
	delete roadQueue;
	delete gapQueue;
	
	// get the road and gap queue from the road start
	roadQueue = theRoadStart->getRoadQueue();
	gapQueue = theRoadStart->getGapQueue();

	// remeber that we are connected now
	theRoadStart->markDirectlyConnectedTo(this);
	partnerRoadStart = theRoadStart;
	directyConnected = true;
	nextSyncTime = numeric_limits<double>::max();
}


unsigned RoadEnd::getNumberOfAgentsOnRoad() const
{
	return roadQueue->size();
}

void RoadEnd::printAgentOnRoad() const
{
	if(!roadQueue->empty())
		cout << "on road " << id << " agent " << roadQueue->front()->getId() << " is still waiting" << endl;
}

void RoadEnd::notifyFirstCarAboutDepartureTime()
{
	if(!roadQueue->empty() && !firstCarNotified)
	{
		// inform the first car in the row to add a timer to the system for its departure time
		firstCarNotified = true;
		roadQueue->front()->handleEarliestTimeToLeaveRoad(lastLeavingTime + minimalGap);
	}
}

void RoadEnd::createShortExchangeMessage()
{
	// this is to be used if the road end is on the same process
	exchangeMessageSend.clear();
	Serializer::setVector(&exchangeMessageSend);
	Serializer::serializeToVector(0U);
	Serializer::serializeToVector(static_cast<unsigned>(gapQueue->size()));
	for(list<double>::const_iterator queueIterator = gapQueue->begin(); 
	    queueIterator != gapQueue->end(); ++queueIterator)
		Serializer::serializeToVector(*queueIterator);
}

void RoadEnd::createLongExchangeMessage()
{
	// this is to be used if the road end is on a different process
	exchangeMessageSend.clear();
	Serializer::setVector(&exchangeMessageSend);
	Serializer::serializeToVector(1U);
	Serializer::serializeToVector(static_cast<unsigned>(gapQueue->size()));
	for(list<double>::const_iterator queueIterator = gapQueue->begin(); 
	    queueIterator != gapQueue->end(); ++queueIterator)
		Serializer::serializeToVector(*queueIterator);
}

void RoadEnd::sendExchangeMessage()
{
	int sendToProcID = dynamic_cast<SimNode*>(myNetwork.getNodeWithId(fromNodeId))->getProcId();
	Communicator::sendToProcess(sendToProcID, static_cast<int>(id * 2U + 1U), exchangeMessageSend, myClock.getLastTimerTime());
}

void RoadEnd::emptyGapQueue()
{
	gapQueue->clear();
}

void RoadEnd::receiveExchangeMessage()
{
	int recvFromProcID = dynamic_cast<SimNode*>(myNetwork.getNodeWithId(fromNodeId))->getProcId();
	Communicator::receiveFromProcess(recvFromProcID, static_cast<int>(id * 2U), exchangeMessageReceive, myClock.getLastTimerTime());
}

void RoadEnd::readShortExchangeMessage()
{
	// this is to be used if the road end is on the same process
	unsigned messageType, roadQueueSize;
	Agent *agent;

	vector<char>::const_iterator messageIterator = exchangeMessageReceive.begin();

	messageIterator = Serializer::unserializeFromVectorIterator(messageType, messageIterator);
	assert(messageType == 0U);
	messageIterator = Serializer::unserializeFromVectorIterator(roadQueueSize, messageIterator);
	for(unsigned i = 0; i < roadQueueSize; ++i)
	{
		messageIterator = Serializer::unserializeFromVectorIterator(agent, messageIterator);
		roadQueue->push_back(agent);
	}
	throw;
}

void RoadEnd::readLongExchangeMessage()
{
	// this is to be used if the road end is on a different process
	unsigned messageType, roadQueueSize;

	vector<char>::const_iterator messageIterator = exchangeMessageReceive.begin();

	messageIterator = Serializer::unserializeFromVectorIterator(messageType, messageIterator);
	if(messageType != 1U)
	{
		Communicator::getLogFile() << "Error in RoadEnd::readLongExchangeMessage(): expected long message but received short message" << endl;
		throw "Error in RoadEnd::readLongExchangeMessage(): expected long message but received short message";
	}
	messageIterator = Serializer::unserializeFromVectorIterator(roadQueueSize, messageIterator);
	for(unsigned i = 0; i < roadQueueSize; ++i)
	{
		unsigned agentId;
		messageIterator = Serializer::unserializeFromVectorIterator(agentId, messageIterator);
		// make sure this agent does not yet exist...
		assert(!myWorld.hasAgent(agentId));
		// create a new agent, unserialize it and add it to road queue and world
		Agent *newAgent = new Agent(agentId, myWorld);
		messageIterator = newAgent->unserializeFromVectorIterator(messageIterator);
		roadQueue->push_back(newAgent);
		myWorld.addAgent(newAgent);
	}
}


}
}
