/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "world.h"
#include "network.h"
#include "clock.h"
#include "simnode.h"
#include "link.h"
#include "roadstart.h"
#include "roadend.h"
#include "agent.h"
#include "serializer.h"
#include "communicator.h"
#include <cassert>
#include <sstream>
#include <iostream>
#include <cmath>

namespace allsim {
namespace microsim {
using namespace std;

RoadStart::RoadStart(unsigned id, unsigned fromNodeId, unsigned toNodeId, 
                     double freeSpeedTravelTime, double gapTravelTime, 
                     double minimalGap, unsigned sizeOfRoad,
                     double squeezeTime, World &theWorld)
:id(id),
fromNodeId(fromNodeId),
toNodeId(toNodeId),
freeSpeedTravelTime(freeSpeedTravelTime),
syncTime(min(freeSpeedTravelTime, gapTravelTime)),
minimalGap(minimalGap),
squeezeTime(squeezeTime),
myWorld(theWorld),
myNetwork(*theWorld.getNetwork()),
myClock(theWorld.getClock()),
numGapsDirectlyAvailable(max(sizeOfRoad, 1u)),
firstWaitingCarHasGap(false),
firstWaitingCarHasSqueezeTimer(false),
lastEntryTime(-numeric_limits<double>::max()),
lastRegisterTime(-numeric_limits<double>::max()),
nextSyncTime(theWorld.getFirstSyncTime()),
//nextSyncTime(numeric_limits<double>::max()),
isInSync(false),
directyConnected(false),
partnerRoadEnd(NULL)
{
// 	Communicator::getLogFile() << "rs: "
// 	<< "id=" << id << " fstt=" << freeSpeedTravelTime << " gtt=" << gapTravelTime 
// 	<< " cap=" << 1/minimalGap << " mg=" << minimalGap << " size=" << sizeOfRoad 
// 	<< " sqt=" << squeezeTime << " synct=" << syncTime << endl;
}


RoadStart::~RoadStart()
{
}

unsigned RoadStart::getId()
{
	return id;
}


void RoadStart::registerAgentForEntering(Agent *agent, double time)
{
	if(DEBUG && !ENTRY_ONLY && (ALL || id == LINKID1 || id == LINKID2 || agent->getId() == AGENTID1 || agent->getId() == AGENTID2))
	{
		cout.precision(DISP_PREC);
		cout << "r a" << agent->getId() << " r" 
		     << id << " t" << time << endl;
	}
	assert(time >= lastRegisterTime);
	lastRegisterTime = time;
	entryQueue.push_back(agent);
}


void RoadStart::registerAgentForStartingATrip(Agent *agent, double time)
{
	// remarks:
	// the starting queue holds all agents that currently would
	// like to start their trip on this road. 
	// It is important not to add them directly to the entry queue.
	// Otherwise, they would have precedence of all cars coming 
	// from the network and wanting to enter this road.
	// the road makes sure that there is always only one car from the 
	// starting queue in the entry queue. This gives the starting
	// queue the same priority as any other regular road.
	// ------------------------------

	if(DEBUG && !ENTRY_ONLY &&(ALL || id == LINKID1 || id == LINKID2 || agent->getId() == AGENTID1 || agent->getId() == AGENTID2))
	{
		cout.precision(DISP_PREC);
		cout << "b a" << agent->getId() << " r" 
		     << id << " t" << time << endl;
	}
	
	// output matsim type event
	if(WRITE_MATSIM_EVENTS)
	{
		if(Communicator::isEventsFileBinary())
		{
			Communicator::writeDoubleToStream(time, Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(agent->getId(), Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(id, Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(6, Communicator::getEventsFile()); // start trip
		}
		else
		{
			Communicator::getEventsFile() << static_cast<int>(floor(time+0.5)) << "\t" 
			                              << agent->getId() << "\t" 
			                              << agent->getCurrentTripIndex() << "\t" << id << "\t"
			                              << myNetwork.getLinkWithId(id)->getFromNode()->getId() 
			                              << "\t" << "6\tstarting\n";
		}
	}

	// add the agent to the starting queue
	if(time < lastRegisterTime || time > 1e10)
	{
		cerr << "error, wrong register time " << endl
		     << "agentid=" << agent->getId()
		     << " time=" << time << " lastRegisterTime=" << lastRegisterTime
		     << " lastEntryTime=" << lastEntryTime << endl;
		cerr << "this link id is " << id << endl;
		throw;
	}
	assert(time >= lastRegisterTime);
	lastRegisterTime = time;

	startingQueue.push_back(agent);
	// if this agent is the first one in the starting queue add
	// it also to the entry queue
	if(startingQueue.size() == 1)
		entryQueue.push_back(agent);
}

void RoadStart::noteAgentEntersRoad(Agent *agent, double time)
{
	// count this agent on the from node
	if(COMPUTE_LOADS)
	{
		SimNode *fromNode = dynamic_cast<SimNode*>(myNetwork.getNodeWithId(fromNodeId));
		if(fromNode)
			fromNode->increaseOutVolume();
		else
			throw "Cast error in RoadStart::noteAgentEntersRoad()";
	}
	

	if(DEBUG && (ALL || id == LINKID1 || id == LINKID2 || agent->getId() == AGENTID1 || agent->getId() == AGENTID2))
	{
		cout.precision(DISP_PREC);
		cout << "e a" << agent->getId() << " r" 
		     << id << " t" << time << " " << numGapsDirectlyAvailable + gapQueue.size() << endl;
	}

	

	if(entryQueue.front() == agent)
	{
		double epsilon = 1e-10;
		if(time < lastEntryTime + minimalGap - epsilon || time < agent->getLastRegisterTime() - epsilon)
		{
			cerr.precision(30);
			cerr << "error, agent tried to enter road at a time too early" << endl;
			cerr << "agent " << agent->getId() << ", letpmg=" << lastEntryTime+minimalGap << ", ag reg time=" << agent->getLastRegisterTime() << ", time= " << time << endl;
			cerr << "time < lastEntryTime + minimalGap = " << (time < lastEntryTime + minimalGap) << endl;
			cerr << "time - (lastEntryTime + minimalGap) = " << (time - (lastEntryTime + minimalGap)) << endl; 
			cerr << "(time - minimalGap) - lastEntryTime = " << ((time - minimalGap) - lastEntryTime) << endl; 
			cerr << "time < agent->getLastRegisterTime() = " << (time < agent->getLastRegisterTime()) << endl;
			throw;
		}
		// store the time of this entry
		lastEntryTime = max(lastEntryTime, time);

		// handling of entering cars...
		// if this agent comes from the starting queue
		// remove it from there and add a new agent from it to
		// the entry queue
		if(!startingQueue.empty() && agent == startingQueue.front())
		{
			// write also a matsim event for the starting car
			if(WRITE_MATSIM_EVENTS)
				if(Communicator::isEventsFileBinary())
				{
					Communicator::writeDoubleToStream(time, Communicator::getEventsFile());
					Communicator::write32BitIntegerToStream(agent->getId(), Communicator::getEventsFile());
					Communicator::write32BitIntegerToStream(id, Communicator::getEventsFile());
					Communicator::write32BitIntegerToStream(4, Communicator::getEventsFile()); // enter network
				}
				else
				{
					Communicator::getEventsFile() 
						<< static_cast<int>(floor(time+0.5)) << "\t" 
						<< agent->getId() << "\t" 
						<< agent->getCurrentTripIndex() << "\t" 
						<< id << "\t"
						<< myNetwork.getLinkWithId(id)->getFromNode()->getId() 
						<< "\t" << "4\tenter net\n";
				}
			startingQueue.pop_front();
			if(!startingQueue.empty())
				entryQueue.push_back(startingQueue.front());
		}
			// write also a matsim event for the entering car
		if(WRITE_MATSIM_EVENTS)
		{
			if(Communicator::isEventsFileBinary())
			{
				Communicator::writeDoubleToStream(time, Communicator::getEventsFile());
				Communicator::write32BitIntegerToStream(agent->getId(), Communicator::getEventsFile());
				Communicator::write32BitIntegerToStream(id, Communicator::getEventsFile());
				Communicator::write32BitIntegerToStream(5, Communicator::getEventsFile()); // enter link
			}
			else
			{
				Communicator::getEventsFile() 
					<< static_cast<int>(floor(time+0.5)) << "\t" 
					<< agent->getId() << "\t" 
					<< agent->getCurrentTripIndex() << "\t" 
					<< id << "\t"
					<< myNetwork.getLinkWithId(id)->getFromNode()->getId() 
					<< "\t" << "5\tenter\n";
			}
		}

		// remove car from entry queue
		entryQueue.pop_front();
		// add car to the road queue
		roadQueue.push_back(agent);
		// ... and remove the corresponding gap (if there is one)
		if(numGapsDirectlyAvailable > 0)
		{
			--numGapsDirectlyAvailable;
			agent->setSqueezeFlag(false);
		}
		else
		{
			// it there is a gap until time, the car uses it
			// if there is none, the care is squeezing in.
			double earliestGap = numeric_limits<double>::max();
			if(!gapQueue.empty())
				earliestGap = gapQueue.front();
			if(time >= earliestGap)
			{
				gapQueue.pop_front();
				agent->setSqueezeFlag(false);
			}
			else
			{
				// this car is squeezing in
				agent->setSqueezeFlag(true);
			}
		}
		// since the first car in the road queue was processed the current first car has not yet received a gap.
		firstWaitingCarHasGap = false;
		firstWaitingCarHasSqueezeTimer = false;
		// invalidate any squeeze timer
		invalidateNewestSqueezeTimer();
	}
	else
	{
		ostringstream errorString;
		errorString << "Error: Agent " << agent->getId() << 
		               " tried to enter road " << id << " without registering in advance.";
		throw(errorString.str());
	}
	// "normalize" gap queue -> remove gaps that arrived before the lastEntryTime and instead increase
	// the number of directly available gaps
	while(!gapQueue.empty())
		if(gapQueue.front() <= lastEntryTime + minimalGap)
		{
			++numGapsDirectlyAvailable;
			gapQueue.pop_front();
		}
		else
			break;
}

void RoadStart::noteAgentEndsTrip(Agent *agent, double time)
{
	if(DEBUG && !ENTRY_ONLY && (ALL || id == LINKID1 || id == LINKID2 || agent->getId() == AGENTID1 || agent->getId() == AGENTID2))
	{
		cout.precision(DISP_PREC);
		cout << "y a" << agent->getId() << " r" 
		     << id << " t" << time << endl;
	}
	// output matsim style events...
	if(WRITE_MATSIM_EVENTS)
	{
		if(Communicator::isEventsFileBinary())
		{
			Communicator::writeDoubleToStream(time, Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(agent->getId(), Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(id, Communicator::getEventsFile());
			Communicator::write32BitIntegerToStream(0, Communicator::getEventsFile()); // leave link
		}
		else
		{
			Communicator::getEventsFile() 
				<< static_cast<int>(floor(time+0.5)) << "\t" 
				<< agent->getId() << "\t" 
				<< agent->getCurrentTripIndex() << "\t" 
				<< id << "\t"
				<< myNetwork.getLinkWithId(id)->getFromNode()->getId() 
				<< "\t" << "0\tend\n";
		}
	}


	if(entryQueue.front() == agent)
	{
		// store the time of this entry
		lastEntryTime = max(lastEntryTime, time);

		// handling of entering cars...
		// if this agent comes from the starting queue
		// remove it from there and add a new agent from it to
		// the entry queue
		if(!startingQueue.empty() && agent == startingQueue.front())
		{
			startingQueue.pop_front();
			if(!startingQueue.empty())
				entryQueue.push_back(startingQueue.front());
		}

		// remove car from entry queue
		entryQueue.pop_front();

		// since this car is not really entering the road, we dont add it to the road queue
		// ... and we do not remove the corresponding gap as well 
		// since the first car in the road queue was processed the current first car has not yet received a gap.
		firstWaitingCarHasGap = false;
		firstWaitingCarHasSqueezeTimer = false;
		// mark any squeeze timer as invalid
		invalidateNewestSqueezeTimer();
	}
	else
	{
		ostringstream errorString;
		errorString << "Error: Agent " << agent->getId() << 
		               " tried to enter its trip on road " << id << " without registering in advance.";
		throw(errorString.str());
	}
	// "normalize" gap queue -> remove gaps that arrived before the lastEntryTime and instead increase
	// the number of directly available gaps
	while(!gapQueue.empty())
		if(gapQueue.front() <= lastEntryTime + minimalGap)
		{
			++numGapsDirectlyAvailable;
			gapQueue.pop_front();
		}
		else
			break;
}


double RoadStart::getFreeSpeedLeavingTime(double entryTime) const
{
	return entryTime + freeSpeedTravelTime;
}


void RoadStart::checkTasks()
{
	//	we have to account for other waiting cars that would like to enter the queue...
	findGapForFirstWaitingCar();
	//  if we are directly connected to a road end this has to be informed as well...
	if(directyConnected)
		partnerRoadEnd->checkTasksFromPartner();
}

void RoadStart::checkTasksFromPartner()
{
	//	we have to account for other waiting cars that would like to enter the queue...
	findGapForFirstWaitingCar();
}

void RoadStart::timerExpired()
{
	// this may have two reasons:
	// either a squeeze timer or a sync timer has expired
	if(nextSyncTime < getNextSqueezeTimerTime())
	{
		// debug:
		if(!(fabs(nextSyncTime - myClock.getLastTimerTime()) < 1.1 * SYNC_OFFSET))
		{
			cerr << "roadid = " << id << endl;
			cerr << "nextSyncTime=" << nextSyncTime << ", lastTimer=" << myClock.getLastTimerTime() << endl;
			cerr << "nextSyncTime - lastTimer = " << nextSyncTime - myClock.getLastTimerTime() << endl;
			cerr << "(nextSyncTime - SYNC_OFFSET) - lastTimer = " << (nextSyncTime - SYNC_OFFSET) - myClock.getLastTimerTime() << endl;
			throw;
		}
		// means: this was a sync timer
		// first check, if we are directly connected to a road end. If so, don't do anything
		if(!directyConnected)
		{
			// exchange data with road end and register a new timer with the clock
			exchangeDataWithRoadEnd();
			nextSyncTime += syncTime;
			myClock.registerTimer(this, nextSyncTime);
			checkTasks();

// 			if(isInSync)
// 			{
// 				exchangeDataWithRoadEndReceive();
// 				nextSyncTime += syncTime;
// 
// 				myClock.registerTimer(this, nextSyncTime - SYNC_OFFSET);
// 				myClock.registerTimer(this, nextSyncTime);
// 				isInSync = false;
// 				checkTasks();
// 			}
// 			else
// 			{
// 				isInSync = true;
// 				exchangeDataWithRoadEndSend();
// 			}
		}
		else
		{
			// we should not get a sync timer if we are directly
			// connected !
			throw;
/*			if(isInSync)
			{
				nextSyncTime = numeric_limits<double>::max();
				isInSync = false;
			}
			else
				isInSync = true;*/
		}
	}
	else
	{
		// debug:
		if(DEBUG && SQUEEZE && !ENTRY_ONLY && (ALL || id == LINKID1 || id == LINKID2 || entryQueue.front()->getId() == AGENTID1 || entryQueue.front()->getId() == AGENTID2) )
		{
			cout << "t a" << entryQueue.front()->getId() << " r" << id 
			     << " t" << getNextSqueezeTimerTime() 
			     << " valid=" << isNextSqueezeTimerValid() << endl;
		}
		// means: it was a squeezetimer that expired.
		if(isNextSqueezeTimerValid())
		{
			if(DEBUG && !ENTRY_ONLY && (ALL || id == LINKID1 || id == LINKID2 || entryQueue.front()->getId() == AGENTID1 || entryQueue.front()->getId() == AGENTID2))
			{
				cout << "q a" << entryQueue.front()->getId() << " r" << id << " t" << getNextSqueezeTimerTime() << endl;
			}
			assert(!firstWaitingCarHasGap);
			assert(squeezeTimers.size() == 1);

			// we squeeze the car in
			assert(getNextSqueezeTimerTime() == myClock.getLastTimerTime());

			firstWaitingCarHasGap = true;
			double tmp = getNextSqueezeTimerTime();
			removeNextSqueezeTimer();
			entryQueue.front()->handleTimeToEnterRoad(tmp);	
		}
		else
		{
			if(DEBUG && SQUEEZE && !ENTRY_ONLY && (ALL || id == LINKID1 || id == LINKID2 || entryQueue.front()->getId() == AGENTID1 || entryQueue.front()->getId() == AGENTID2))
 			{
				cout << "i a" << entryQueue.front()->getId() << " r" 
				     << id << " t" << getNextSqueezeTimerTime() << endl;
 			}
			removeNextSqueezeTimer();
			checkTasks();
		}
	}
}

int RoadStart::getSortPriority() const
{
	return 2 * static_cast<int>(id);
}


void RoadStart::registerInitialTimerAtClock()
{
	if(!directyConnected)
	{
		nextSyncTime += syncTime;
//		myClock.registerTimer(this, nextSyncTime-SYNC_OFFSET);
		myClock.registerTimer(this, nextSyncTime);
	}
}

void RoadStart::exchangeDataWithRoadEnd()
{
	// RoadStarts first send and then receive messages

	// send information about cars that entered road start
	createLongExchangeMessage();
	deleteAgentsOnRoadFromQueueAndWorld();
	sendExchangeMessage();
	exchangeMessageSend.clear();

	// receive information about gaps that are travelling backwards through the link
	receiveExchangeMessage();
	readLongExchangeMessage();
	exchangeMessageReceive.clear();
}

void RoadStart::exchangeDataWithRoadEndSend()
{
	// this method is no longer used
	abort();

	createLongExchangeMessage();
	deleteAgentsOnRoadFromQueueAndWorld();
	sendExchangeMessage();
}

void RoadStart::exchangeDataWithRoadEndReceive()
{
	// this method is no longer used
	abort();

	receiveExchangeMessage();
	readLongExchangeMessage();
	exchangeMessageSend.clear();
	exchangeMessageReceive.clear();
}

list<Agent*> *RoadStart::getRoadQueue()
{
	return &roadQueue;
}

list<double> *RoadStart::getGapQueue()
{
	return &gapQueue;
}

void RoadStart::markDirectlyConnectedTo(RoadEnd *roadEnd)
{
	partnerRoadEnd = roadEnd;
	directyConnected = true;
	nextSyncTime = numeric_limits<double>::max();
}

unsigned RoadStart::getNumberOfAgentsOnRoad() const
{
	return roadQueue.size();
}

void RoadStart::printAgentOnRoad() const
{
	if(!roadQueue.empty())
		cout << "on road " << id << " agent " << roadQueue.front()->getId() << " is still waiting" << endl;
}

void RoadStart::findGapForFirstWaitingCar()
{
	// if the entry queue is empty or the first waiting car has already a gap don't do anything
	if(entryQueue.empty() || firstWaitingCarHasGap )
		return;
	
	// first calculate some times that are usefull
	double earliestEntryTime = lastEntryTime + minimalGap;
	double tRegister = entryQueue.front()->getLastRegisterTime();
	double tLatestEntry = max(earliestEntryTime, tRegister) + squeezeTime;
	double earliestGap;

	if(numGapsDirectlyAvailable > 0)
		earliestGap = -numeric_limits<double>::max();
	else
		if(!gapQueue.empty())
			earliestGap = gapQueue.front();
		else
			earliestGap = numeric_limits<double>::max();

	// the time for entering the road queue has to obey these relations
	// t_entry >= t_earliestEntry   <---- capacity constraint
	// t_entry >= t_earliestGap     <---- spacial constraint (car needs a gap to be available)
	// t_entry >= t_register        <---- we cannot enter before requesting it
	// t_entry <= t_latestEntry     <---- agents are impatient and squeeze themselves in.

	double entryTime = max(max(earliestEntryTime, tRegister), earliestGap);

	if(entryTime < tLatestEntry)
	{
		// means: we have a regular entry time -> inform the agent
		firstWaitingCarHasGap = true;
		assert(entryTime >= myClock.getLastTimerTime());
		entryQueue.front()->handleTimeToEnterRoad(entryTime);
	}
	else
	{
		if(firstWaitingCarHasSqueezeTimer == false)
		{
	
			// means: the road is full, but the agent is going to squeeze itsef in
			// register a squeeze timer
			if(DEBUG && SQUEEZE && !ENTRY_ONLY && (ALL || id == LINKID1 || id == LINKID2 || entryQueue.front()->getId() == AGENTID1 || entryQueue.front()->getId() == AGENTID2))
 			{
				cout << "c a" << entryQueue.front()->getId() << " r" 
				     << id << " t" << tLatestEntry << endl;
			}
			assert(tLatestEntry >= myClock.getLastTimerTime());
			addSqueezeTimer(tLatestEntry);
			firstWaitingCarHasSqueezeTimer =  true;
			myClock.registerTimer(this, tLatestEntry);
		}
	}


/*
	// calculate the time when the agent squeezes himself into the road
	double squeezeEntryTime = lastEntryTime + minimalGap + squeezeTime;

	if(numGapsDirectlyAvailable > 0)
	{
		// calculate the regular entry time based on a directly available gap
		double regularEntryTime = max(lastEntryTime + minimalGap, lastRegisterTime);
		firstWaitingCarHasGap = true;
		// debug:
		if(Clock::getLastTimerTime() > regularEntryTime)
		{
			cerr << "warning: lastTimerTime > regularEntryTime" << endl
			     << "lastTimerTime=" << Clock::getLastTimerTime()
			     << ", regularEntryTime=" << regularEntryTime << endl
			     << "roadid = " << id << endl;
		}
		entryQueue.front()->handleTimeToEnterRoad(regularEntryTime);	
	}
	else
	{
		if(!gapQueue.empty() && (gapQueue.front() < squeezeEntryTime))
		{
			// calculate the regular entry time based on a gap that will open in the future
			double regularEntryTime = max(gapQueue.front(), lastEntryTime + minimalGap);
			regularEntryTime = max(regularEntryTime, lastRegisterTime);
			firstWaitingCarHasGap = true;
			// debug:
			if(Clock::getLastTimerTime() > regularEntryTime)
			{
				cerr << "warning: lastTimerTime > regularEntryTime" << endl
				     << "lastTimerTime=" << Clock::getLastTimerTime()
				     << ", regularEntryTime=" << regularEntryTime << endl
				     << "roadid = " << id << endl;
			}
			entryQueue.front()->handleTimeToEnterRoad(regularEntryTime);	
		}
		else
		{
			// means: there is currently no space left for the agent
			// this makes it necessary to prepare for it to squeeze in
			// -> the road registers a squeeze in timer
			
			// if a squeeze timer already exists, dont register anything
			// only if there was no squeezetimer registered so far register one
			if(nextSqueezeTime == numeric_limits<double>::max())
			{
				nextSqueezeTime = max(squeezeEntryTime, lastRegisterTime);
				squeezeTimerAgent = entryQueue.front();
 				if(id == LINKID1)
					cerr << "registering squeeze timer at " << nextSqueezeTime << " on road " << id  
					     << " *******************************" << endl;
				Clock::registerTimer(this, nextSqueezeTime);
			}
			else
			{
				// dont do anything, since we will get informed early enough
			}
		}
	}*/
}

void RoadStart::addSqueezeTimer(double time)
{
	invalidateNewestSqueezeTimer();
	squeezeTimers.push_back(SqueezeTimer(time));
}	

double RoadStart::getNextSqueezeTimerTime()
{
	if(!squeezeTimers.empty())
		return squeezeTimers.begin()->time;
	else
		return numeric_limits<double>::max();
}

bool RoadStart::isNextSqueezeTimerValid()
{
	if(!squeezeTimers.empty())
		return squeezeTimers.begin()->valid;
	else
		return false;
}

void RoadStart::removeNextSqueezeTimer()
{
	if(!squeezeTimers.empty())
		squeezeTimers.pop_front();
}

void RoadStart::invalidateNewestSqueezeTimer()
{
	if(!squeezeTimers.empty())
	{
		list<SqueezeTimer>::iterator last = squeezeTimers.end();
		--last;
		last->valid = false;
	}
}




void RoadStart::createShortExchangeMessage()
{
	// this is to be used if the road end is on the same process
	exchangeMessageSend.clear();
	Serializer::setVector(&exchangeMessageSend);
	Serializer::serializeToVector(0U);
	Serializer::serializeToVector(static_cast<unsigned>(roadQueue.size()));
	for(list<Agent*>::const_iterator queueIterator = roadQueue.begin(); 
	    queueIterator != roadQueue.end(); ++queueIterator)
		Serializer::serializeToVector(*queueIterator);
}

void RoadStart::createLongExchangeMessage()
{
	// this is to be used if the road end is on a different process
	exchangeMessageSend.clear();
	Serializer::setVector(&exchangeMessageSend);
	Serializer::serializeToVector(1U);
	Serializer::serializeToVector(static_cast<unsigned>(roadQueue.size()));
	for(list<Agent*>::const_iterator queueIterator = roadQueue.begin(); 
	    queueIterator != roadQueue.end(); ++queueIterator)
		(*queueIterator)->serializeToVector(&exchangeMessageSend);
}

void RoadStart::deleteAgentsOnRoadFromQueueAndWorld()
{
	for(list<Agent*>::const_iterator queueIterator = roadQueue.begin(); 
	    queueIterator != roadQueue.end(); ++queueIterator)
	{
		Agent * currentAgent = *queueIterator;
		myWorld.removeAgent(currentAgent->getId());
		delete currentAgent;
	}	
	roadQueue.clear();
}


void RoadStart::sendExchangeMessage()
{
	int sendToProcID = dynamic_cast<SimNode*>(myNetwork.getNodeWithId(toNodeId))->getProcId();
	Communicator::sendToProcess(sendToProcID, static_cast<int>(id * 2U), exchangeMessageSend, myClock.getLastTimerTime());
}

void RoadStart::receiveExchangeMessage()
{
	int recvFromProcID = dynamic_cast<SimNode*>(myNetwork.getNodeWithId(toNodeId))->getProcId();
	Communicator::receiveFromProcess(recvFromProcID, static_cast<int>(id * 2U + 1U), exchangeMessageReceive, myClock.getLastTimerTime());
}

void RoadStart::readShortExchangeMessage()
{
	// this is to be used if the road end is on the same process
	unsigned messageType, gapQueueSize;
	double gapTime;

	vector<char>::const_iterator messageIterator = exchangeMessageReceive.begin();

	messageIterator = Serializer::unserializeFromVectorIterator(messageType, messageIterator);
	assert(messageType == 0U);
	messageIterator = Serializer::unserializeFromVectorIterator(gapQueueSize, messageIterator);
	for(unsigned i = 0; i < gapQueueSize; ++i)
	{
		messageIterator = Serializer::unserializeFromVectorIterator(gapTime, messageIterator);
		gapQueue.push_back(gapTime);
	}
	throw;
}

void RoadStart::readLongExchangeMessage()
{
	// this is to be used if the road end is on a different process
	unsigned messageType, gapQueueSize;
	double gapTime;

	vector<char>::const_iterator messageIterator = exchangeMessageReceive.begin();

	messageIterator = Serializer::unserializeFromVectorIterator(messageType, messageIterator);
	assert(messageType == 1U);
	messageIterator = Serializer::unserializeFromVectorIterator(gapQueueSize, messageIterator);
	for(unsigned i = 0; i < gapQueueSize; ++i)
	{
		messageIterator = Serializer::unserializeFromVectorIterator(gapTime, messageIterator);
		gapQueue.push_back(gapTime);
	}
}




}
}
