/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "simnetwork.h"
#include "simnode.h"
#include "link.h"
#include "roadstart.h"
#include "roadend.h"
#include "agent.h"
#include "serializer.h"
#include "communicator.h"
#include "microsimconfig.h"
#include "global.h"
#include <iostream>
#include <sstream>
#include <algorithm>

namespace allsim {
namespace microsim {

/// \class SimNetwork
/// \brief Network used for the simulation run
/// \author David Charypar

/// Quasi type conversion constructor
/// 
/// A normal network is "augmented" to a simulation network.
SimNetwork::SimNetwork(Network const &baseNet) // copy constructor
:Network(baseNet)
{
	// here, the sim network still contains ordinary Nodes instead of SimNodes
	// we have to change that
	vector<Node*> oldNodes;
	for(map<unsigned, Node*>::const_iterator ni = nodes.begin(); 
	    ni != nodes.end(); ++ni)
	{
		Node *oldNode = ni->second;
		SimNode *newNode = new SimNode(*oldNode);

		nodes[newNode->getId()] = newNode;
		// store old node for later deletion
		oldNodes.push_back(oldNode);
	}
	// we have to exchange the old links with new links that point to the new sim nodes
	for(map<unsigned, Link*>::const_iterator li = links.begin(); 
	    li != links.end(); ++li)
	{
		Link *csl = li->second;
		Node *newFromNode = nodes[csl->getFromNode()->getId()];
		Node *newToNode = nodes[csl->getToNode()->getId()];
		csl->setFromNode(newFromNode);
		csl->setToNode(newToNode);
		// register the link with the new to and from nodes
		newFromNode->addOutgoingLink(csl);
		newToNode->addIncomingLink(csl);
	}
	// delete the old nodes as they are obsolete now
	
	for(vector<Node*>::const_iterator ni = oldNodes.begin();
	    ni != oldNodes.end(); ++ni)
		delete *ni;
}

SimNetwork::~SimNetwork()
{
}

/// Partitions the network using orthogonal recursive bisection
//
/// The bisection algorithm recursively cuts the longest dimension.
/// Subsequent bisection steps dont have to be necessarily orthogonal.
/// The number of bisection steps is given in \b numBisectionSteps. 
/// The resulting network is partitioned into 2^numBisectionSteps parts
void SimNetwork::partitionNetwork(unsigned numBisectionSteps)
{
	vector<SimNode*> nodeVector;
	nodeVector.reserve(nodes.size());
	for(map<unsigned, Node*>::const_iterator nI = nodes.begin();
	    nI != nodes.end(); nI++)
		nodeVector.push_back(dynamic_cast<SimNode*>(nI->second));
	bisectNodes(nodeVector, 0, numBisectionSteps);
//	assignNodesToProc(numWidth, numHeight);
}


void SimNetwork::createRoadsForWorld(World &theWorld)
{
	// cache data from the config file for quick access
	using namespace config;
	double squeezeTime = convertTimeStringToTime( Config::getModuleConfig(Config::DEQSIM).getParam(Config::DEQSIM_SQUEEZETIME));
	
	double flowCapacityFactor = atof(Config::getModuleConfig(Config::DEQSIM).
	                            getParam(Config::DEQSIM_FLOWCAPACITYFACTOR).c_str());
	double spaceCapacityFactor = atof(Config::getModuleConfig(Config::DEQSIM).
	                             getParam(Config::DEQSIM_STORAGECAPACITYFACTOR).c_str());
	double carSize = atof(Config::getModuleConfig(Config::DEQSIM).
	                 getParam(Config::DEQSIM_CARSIZE).c_str());
	double gapTravelSpeed = atof(Config::getModuleConfig(Config::DEQSIM).
	                        getParam(Config::DEQSIM_GAPTRAVELSPEED).c_str());

	if(DEBUG)
	{
		Communicator::getLogFile() << "Used config data:" << endl
			<< "squeezeTime = " << squeezeTime << endl
			<< "flowCapacityFactor = " << flowCapacityFactor << endl
			<< "spaceCapacityFactor = " << spaceCapacityFactor << endl
			<< "carSize = " << carSize << endl
			<< "gapTravelSpeed = " << gapTravelSpeed << endl;
	}



	for(map<unsigned, Link*>::const_iterator li = links.begin(); 
	    li != links.end(); ++li)
	{
		Link *cL = li->second;


		

		// compute flow capacities, storage capacity, and gap travel time
		// Note1: inlet capacity is higher than outlet as it should represent the 'physical'
		// capacity of the link. Often what is called (outflow)-capacity of a link is really
		// the maximal observed outflow volume during a day and therefore influenced by traffic
		// signals, intersection design, and other factors.
		// Note2: Remember: Capacities are measured in cars per second (i.e. not hours)

		double inletFlowCapacity  = flowCapacityFactor * max(cL->getFlowCapacity() * 1.1, 0.5 * cL->getNumLanes());
		double outletFlowCapacity = flowCapacityFactor * cL->getFlowCapacity();
		int storageCapacity = static_cast<unsigned>(spaceCapacityFactor * cL->getSpace() / carSize + 0.5);
		double gapTravelTime = cL->getLength() / gapTravelSpeed; 

		bool startOnProc = dynamic_cast<SimNode*>(cL->getFromNode())->getProcId() 
		                 == static_cast<unsigned>(Communicator::getProcId());
		bool endOnProc = dynamic_cast<SimNode*>(cL->getToNode())->getProcId() 
		              == static_cast<unsigned>(Communicator::getProcId());

		// check if from node is simulated on this process
		if(startOnProc)
		{
			// create a new road start
			RoadStart *newRoadStart = new RoadStart(cL->getId(), 
			                          cL->getFromNode()->getId(), cL->getToNode()->getId(), 
			                          cL->getFreeSpeedTravelTime(), 
			                          gapTravelTime, 1.0 / inletFlowCapacity, 
			                          storageCapacity, 
			                          squeezeTime, theWorld);
			// add it to World
			theWorld.addRoadStart(newRoadStart);
		}
		if(endOnProc)
		{
			// create a new road end
 			RoadEnd *newRoadEnd = new RoadEnd(cL->getId(), 
			                      cL->getFromNode()->getId(), cL->getToNode()->getId(), 
 			                      cL->getFreeSpeedTravelTime(), 
 			                      gapTravelTime, 1.0 / outletFlowCapacity, 
			                      storageCapacity, 
			                      theWorld);
			// add it to World
			theWorld.addRoadEnd(newRoadEnd);
		}
	}
	
}

/// Prints out accumulated load data too an output stream
void SimNetwork::printNodeLoadsToStream(ostream &outFile)
{
	for(map<unsigned int, Node*>::const_iterator nI = nodes.begin();
	    nI != nodes.end(); ++nI)
	{
		SimNode const *currNode = dynamic_cast<SimNode const *>(nI->second);
		if(currNode)
			outFile << currNode->getId() << "\t" << currNode->getTotalTrafficVolume() << endl;
		else	
			throw "Cast error in SimNetwork::printNodeLoadsToStream()";
	}
}

/// Loads prior load data from a file. This data is later used for orthogonal recursive bisection.
void SimNetwork::loadLoadsFromStream(istream &inFile)
{
	int id, load;
	inFile >> id >> load;
	while(inFile)
	{
		nodeLoads[id] = load;
		inFile >> id >> load;
	}
}

/// Writes the assignment of each link to processors out to a stream. Later used for visualization.
void SimNetwork::printLinkProcAssignmentToStream(ostream &outStream)
{
	for(map<unsigned int, Link*>::const_iterator lI = links.begin();
	    lI != links.end(); ++lI)
	{
		Link const *currLink = lI->second;
		SimNode const *fromNode = dynamic_cast<SimNode const *>(currLink->getFromNode());
		SimNode const *toNode = dynamic_cast<SimNode const *>(currLink->getToNode());
		if((fromNode && toNode))
			outStream << currLink->getId() << "\t" 
			        << fromNode->getProcId() << "\t" 
			        << toNode->getProcId() << endl;
		else	
			throw "Cast error in SimNetwork::printLinkProcAssignmentToStream()";
	}
}



void SimNetwork::assignNodesToProc(unsigned numWidth, unsigned numHeight)
{
	// perform a cartesian decomposition of the network
	// step to complete...
	//1.: find extension of network
	//2.: for all nodes find the node they belong to
	//3.: for all links find the node they belong to
	
	// step 1

	if(nodes.size() <= 0)
		return;

	// find extentsion of the network
	map<unsigned, Node*>::const_iterator iter = nodes.begin();
	Node *currentNode = nodes.begin()->second;
	double minx, maxx, miny, maxy;
	minx = maxx = currentNode->getXCoord();
	miny = maxy = currentNode->getYCoord();

	for(++iter; iter != nodes.end(); ++iter)
	{
		currentNode = iter->second;
		minx = min(minx, currentNode->getXCoord());
		maxx = max(maxx, currentNode->getXCoord());
		miny = min(miny, currentNode->getYCoord());
		maxy = max(maxy, currentNode->getYCoord());
	}
	
	double sizeX = (maxx - minx) * 1.000001;
	double sizeY = (maxy - miny) * 1.000001;
	Communicator::getLogFile() << "sizeX=" << sizeX << ", sizeY=" << sizeY << endl;

	//  step 2
	for(iter = nodes.begin(); iter != nodes.end(); ++iter)
	{
		currentNode = iter->second;
		unsigned nodeIndexX, nodeIndexY;
		if(numWidth == 2)
		{
			if(currentNode->getXCoord()-minx < 900.0e3)
				nodeIndexX = 0;
			else
				nodeIndexX = 1;	
		}
		else
		{
			if(numWidth == 3)
				if(currentNode->getXCoord()-minx < 855.0e3)
					nodeIndexX = 0;
				else
					if(currentNode->getXCoord()-minx < 960.0e3)
						nodeIndexX = 1;	
					else
						nodeIndexX = 2;	
			else
				nodeIndexX = static_cast<unsigned>(numWidth * (currentNode->getXCoord()-minx) / sizeX);
		}
		if(numHeight == 2)
		{
			if(currentNode->getYCoord()-miny < 797.0e3)
				nodeIndexY = 0;
			else
				nodeIndexY = 1;	
		}
		else
		{
			if(numHeight == 3)
				if(currentNode->getYCoord()-miny < 753.0e3)
					nodeIndexY = 0;
				else
					if(currentNode->getYCoord()-miny < 827.0e3)
						nodeIndexY = 1;	
					else
						nodeIndexY = 2;	
			else
				nodeIndexY = static_cast<unsigned>(numHeight * (currentNode->getYCoord()-miny) / sizeY);
		}
		dynamic_cast<SimNode*>(currentNode)->setProcId(numWidth * nodeIndexY + nodeIndexX);
	}
}


void SimNetwork::bisectNodes(vector<SimNode*> &nodesVector, unsigned minId, unsigned remainingBisectionSteps)
{
	ofstream &logFile = Communicator::getLogFile();
	if(DEBUG)
		logFile << "remaining steps = " << remainingBisectionSteps << endl;
	if(remainingBisectionSteps > 0 && nodesVector.size() > 1)
	{
		int cutPointX, cutPointY, cutPoint;		

		// compute horizontal extension and cutpoint in x-Direction
		sort(nodesVector.begin(), nodesVector.end(), ptr_fun(compareXOfTwoNodes));
		double minX = (*nodesVector.begin())->getXCoord();
		double maxX = (*(nodesVector.end()-1))->getXCoord();
		double xExtension = maxX - minX;

		if(DEBUG)
			logFile << "xExtension = " << xExtension << endl;

		if(LOAD_PARTITIONING == true)
		{
			cutPointX = getLoadBasedCutPoint(nodesVector.begin(), nodesVector.end());
		}
		else
		{
			cutPointX = nodesVector.size() / 2;
		}
		double cpX = (*(nodesVector.begin() + cutPointX))->getXCoord();

		// compute vertical extension  and the cutpoitn in y-Direction
		sort(nodesVector.begin(), nodesVector.end(), ptr_fun(compareYOfTwoNodes));
		double minY = (*nodesVector.begin())->getYCoord();
		double maxY = (*(nodesVector.end()-1))->getYCoord();
		double yExtension = maxY - minY;

		if(DEBUG)
			logFile << "yExtension = " << yExtension << endl;

		if(LOAD_PARTITIONING == true)
		{
			cutPointY = getLoadBasedCutPoint(nodesVector.begin(), nodesVector.end());
		}
		else
		{
			cutPointY = nodesVector.size() / 2;
		}
		double cpY = (*(nodesVector.begin() + cutPointY))->getYCoord();

		// compute the aspect ratios of the resulting parts for bothe split directions individually
		// and select the direction that produces a better result
	
		double x1 = maxX - cpX; double x2 = cpX - minX;
		double qualityX = min((min(x1, yExtension) / max(x1, yExtension)), (min(x2, yExtension) / max(x2, yExtension)));

		double y1 = maxY - cpY; double y2 = cpY - minY;
		double qualityY = min((min(y1, xExtension) / max(y1, xExtension)), (min(y2, xExtension) / max(y2, xExtension)));

		if(DEBUG)
			logFile << "Quality-X = " << qualityX << ", Quality-Y = " << qualityY << endl;

		// split along the longer extension
		if(qualityX > qualityY)
		{
			// need to sort again
			sort(nodesVector.begin(), nodesVector.end(), ptr_fun(compareXOfTwoNodes));
			cutPoint = cutPointX;
			if(DEBUG)
				logFile << "partitioning along x-Axis" << endl;
		}
		else
		{
			// the vector should be already sorted along the y-Axis.
			// We include this line for symmetry reasons
			sort(nodesVector.begin(), nodesVector.end(), ptr_fun(compareYOfTwoNodes));
			cutPoint = cutPointY;
			if(DEBUG)
				logFile << "partitioning along y-Axis" << endl;
		}

		// make two partial copies of the vector for the first and second part
		vector<SimNode*> firstPart(nodesVector.begin(), nodesVector.begin()+cutPoint);
		vector<SimNode*> secondPart(nodesVector.begin()+cutPoint, nodesVector.end());

		// call this method recursively
		bisectNodes(firstPart, minId, remainingBisectionSteps - 1);
		bisectNodes(secondPart, minId + (1u << (remainingBisectionSteps - 1)),
		            remainingBisectionSteps - 1);
	}
	else
	{
		if(DEBUG)
			logFile << "assinging " << nodesVector.size() 
			     << " nodes to proc " << minId << endl;
		// set procID of all SimNodes int the nodesVector to minId
		for_each(nodesVector.begin(), nodesVector.end(), 
		         bind2nd(ptr_fun(setProcIdOfSimNode), minId));
	}
}

int SimNetwork::getLoadBasedCutPoint(vector<SimNode*>::const_iterator iStart,
	                             vector<SimNode*>::const_iterator iEnd)
{
	int sumLoad = 0;
	// first compute total traffic load on nodes
	vector<SimNode*>::const_iterator iN;
	for(iN = iStart; iN != iEnd; ++iN)
	{
		sumLoad += nodeLoads[(*iN)->getId()];
	}
	int accumLoad = 0;;
	iN = iStart;
	while((iN != iEnd) && (accumLoad < sumLoad / 2))
	{
		accumLoad += nodeLoads[(*iN)->getId()];
		++iN;
	}
	return iN - iStart;
}


void SimNetwork::setProcIdOfSimNode(SimNode *sn, unsigned newProcId)
{
	sn->setProcId(newProcId);
}

/// Compares the x-coordinate of two Node objects. Returns true if x1 is smaller than x2.
///
/// This static method can be used to sort a container of Node objects
/// according to the x-coordinate of the objects. E.g. it can be used using
/// the \b ptr_fun adaptor of the standard C++ library to call the \b sort algorithm.
bool SimNetwork::compareXOfTwoNodes(Node const *sn1, Node const *sn2)
{
	return sn1->getXCoord() < sn2->getXCoord();
}

/// Compares the y-coordinate of two Node objects. Returns true if y1 is smaller than y2.
///
/// This static method can be used to sort a container of Node objects
/// according to the y-coordinate of the objects. E.g. it can be used using
/// the \b ptr_fun adaptor of the standard C++ library to call the \b sort algorithm.
bool SimNetwork::compareYOfTwoNodes(Node const *sn1, Node const *sn2)
{
	return sn1->getYCoord() < sn2->getYCoord();
}





}
} // end namespace allsim
