/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "global.h"
#include "simnode.h"
#include "link.h"
#include "network.h"
#include "communicator.h"

#include <vector>
#include <iostream>

using namespace std;

namespace allsim {
namespace microsim {

/// \class SimNode
/// \brief SimNode extends the normal Node by adding a processor ID that
/// associates the SimNode with a processor.
/// \author David Charypar

/// Normal constructor. 
///
/// \param xCoord The x-coordinate of this node in the world.
/// \param yCoord The y-coordinate of this node in the world.
/// \param id The ID of the processor this node is assigned to.
SimNode::SimNode(unsigned id, double xCoord, double yCoord)
:Node(id, xCoord, yCoord),
procId(0)
{
}

/// Quasi type conversion constructor that augments an ordinary Node to a SimNode 
/// assigning a default value (0) to procId.
SimNode::SimNode(Node const &node)
:Node(node), procId(0), totalInVolume(0), totalOutVolume(0)
{
}

SimNode::~SimNode()
{
}

/// Returns the ID of the processor which this SimNode object is assigned to.
///
/// This method is used during creation of the SimNetwork to determine
/// which road segment need to communicate over the computer network and which 
/// road segments are completely simulated on a single node.
unsigned SimNode::getProcId() const
{
	return procId;
}

/// Sets the ID of the processor which this SimNode object is assigned to.
///
/// This method is used before creation of the SimNetwork to predefine
/// which road segment need to communicate over the computer network and which 
/// road segments are completely simulated on a single node.
void SimNode::setProcId(unsigned newProcId)
{
	procId = newProcId;
}
/// Returns the total traffic volume that was encountered at this node
unsigned SimNode::getTotalTrafficVolume() const
{
	return totalInVolume + totalOutVolume;
}

/// Used by RoadEnd to register each car that travels in on a link directing to this node
void SimNode::increaseInVolume()
{
	++totalInVolume;
}

/// Used by RoadStart to register each car that travels on a link directing away from this node.
void SimNode::increaseOutVolume()
{
	++totalOutVolume;
}




}
}
