/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "trip.h"
#include "serializer.h"
#include <limits>

namespace allsim {
namespace microsim {

using namespace std;

unsigned Trip::creationCounter = 0;


Trip::Trip(double departureTime, double travelTime)
:departureTime(departureTime), travelTime(travelTime),
route(0),
nextLinkIndex(0)
{
	++creationCounter;
}


Trip::~Trip()
{
}


void Trip::addLinkIdToBackOfRoute(unsigned newLinkId)
{
	route.push_back(newLinkId);
}

void Trip::addLinkIdToFrontOfRoute(unsigned newLinkId)
{
	route.push_back(newLinkId);
	for(unsigned i = route.size()-1; i > 0; --i)
		route[i] = route[i-1];
	route[0] = newLinkId;
}

void Trip::setNextLinkIndex(unsigned nextLinkIndex)
{
	this->nextLinkIndex = nextLinkIndex;
}


double Trip::getDepartureTime() const
{
	return departureTime;
}

double Trip::getTravelTime() const
{
	return travelTime;
}


unsigned Trip::getNextLinkIdAndAdvance()
{
	if(route.size() > nextLinkIndex)
		return route[nextLinkIndex++];
	else
		return numeric_limits<unsigned>::max();
}

unsigned Trip::getStartLinkId() const
{
	if(route.size() > 0)
		return route[0];
	else
		return numeric_limits<unsigned>::max();
}

unsigned Trip::getEndLinkId() const
{
	if(route.size() > 0)
		return route[route.size()-1];
	else
		return numeric_limits<unsigned>::max();
}

unsigned Trip::getSize() const
{
	return route.size();
}

bool Trip::isDone() const
{
	return (nextLinkIndex >= route.size());
}



void Trip::serializeToVector(vector<char> *theVector) const
{
	Serializer::setVector(theVector);

	Serializer::serializeToVector(departureTime);
	Serializer::serializeToVector(travelTime);
	Serializer::serializeToVector(static_cast<unsigned>(route.size()));

	for(vector<unsigned>::const_iterator ui = route.begin();
	                                ui != route.end(); ++ui)
		Serializer::serializeToVector(*ui);

	Serializer::serializeToVector(nextLinkIndex);
}

vector<char>::const_iterator Trip::unserializeFromVectorIterator(Trip*& createdTrip, vector<char>::const_iterator iter)
{
	double departureTime, travelTime;
	unsigned routeSize, roadId, nextIndex;
	iter = Serializer::unserializeFromVectorIterator(departureTime, iter);
	iter = Serializer::unserializeFromVectorIterator(travelTime, iter);
	
	createdTrip = new Trip(departureTime, travelTime);

	iter = Serializer::unserializeFromVectorIterator(routeSize, iter);
	for(unsigned i = 0; i < routeSize; ++i)
	{
		iter = Serializer::unserializeFromVectorIterator(roadId, iter);
		createdTrip->addLinkIdToBackOfRoute(roadId);
	}

	iter = Serializer::unserializeFromVectorIterator(nextIndex, iter);

	createdTrip->setNextLinkIndex(nextIndex);
	

	return iter;		
}


unsigned Trip::getNumCreatedTrips()
{
	return creationCounter;
}

}


}
