/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "world.h"
#include "agent.h"
#include "roadstart.h"
#include "roadend.h"
#include "network.h"
#include "serializer.h"
#include "clock.h"
#include "communicator.h"
#include "microsimconfig.h"
#include <cassert>
#include <iostream>
#include <fstream>
#include <algorithm>


namespace allsim {
namespace microsim {

using namespace config;

World::World()
:firstSyncTime(numeric_limits<double>::max()), numFinishedAgents(0), myClock(*this), theNetwork(NULL)
{
	firstSyncTime = atof(Config::getModuleConfig(Config::DEQSIM).getParam(Config::DEQSIM_STARTTIME).c_str());
}

World::~World()
{
	// the world is responsible for destroying the agents
	for(map<unsigned, RoadStart*>::const_iterator li = roadStarts.begin(); 
	    li != roadStarts.end(); ++li)
	{
		delete li->second;
	}
	roadStarts.clear();

	// the world is responsible for destroying the agents
	for(map<unsigned, RoadEnd*>::const_iterator li = roadEnds.begin(); 
	    li != roadEnds.end(); ++li)
	{
		delete li->second;
	}
	roadEnds.clear();

	for(map<unsigned, Agent*>::const_iterator ai = agents.begin(); 
	    ai != agents.end(); ++ai)
	{
		delete ai->second;
	}
	agents.clear();

}

void World::setNetwork(tools::Network *newNetwork)
{
	theNetwork = newNetwork;
	if(!agents.empty())
		cerr << "warning: setting a new network while agents are already "
		     << "inserted into the world. This can be dangerous..." << endl;
}

tools::Network * World::getNetwork()
{
	assert(theNetwork != NULL);
	return theNetwork;
}

Clock &World::getClock()
{
	return myClock;
}

void World::addAgent(Agent *agent)
{
	if(!hasAgent(agent->getId()))
		agents[agent->getId()] = agent;
	else
	{
		Communicator::getLogFile()
		<< "Error in World::addAgent(): trying to add agent with existing id "
		<< agent->getId() << endl;
		throw "Error in World::addAgent(): trying to add agent with existing id";
	}
}

void World::addRoadStart(RoadStart *roadStart)
{
	if(!hasRoadStart(roadStart->getId()))
		roadStarts[roadStart->getId()] = roadStart;
	else
	{
		Communicator::getLogFile()
		<< "Error in World::addRoadStart(): trying to add RoadStart with existing id "
		<< roadStart->getId() << endl;
		throw "Error in World::addRoadStart(): trying to add RoadStart with existing id";
	}
}

void World::addRoadEnd(RoadEnd *roadEnd)
{
	if(!hasRoadEnd(roadEnd->getId()))
		roadEnds[roadEnd->getId()] = roadEnd;
	else
	{
		Communicator::getLogFile()
		<< "Error in World::addRoadEnd(): trying to add RoadEnd with existing id "
		<< roadEnd->getId() << endl;
		throw "Error in World::addRoadEnd(): trying to add RoadEnd with existing id";
	}
}

void World::removeAgent(unsigned agentId)
{
	if(hasAgent(agentId))
		agents.erase(agentId);
	else
	{
		Communicator::getLogFile()
		<< "Error in World::removeAgent(): trying to remove Agent that does not exist: id="
		<< agentId << endl;
		throw "Error in World::removeAgent(): trying to remove Agent that does not exist";
	}
}

bool World::hasAgent(unsigned agentId)
{
	return agents.find(agentId) != agents.end();
}

Agent *World::getAgent(unsigned agentId)
{
	if(hasAgent(agentId))
		return agents[agentId];
	else
	{
		Communicator::getLogFile() << "Error in World::getAgent(): "
		                           << "Requested Agetn with id " 
		                           << agentId << " which does not exist" << endl;
		throw "World::getAgent(): Requested inexistant Agent";
		return NULL;
	}
}

void World::restartAgentLoop()
{
	currentAgentIt = agents.begin();
}

bool World::hasNextAgent()
{
	return currentAgentIt != agents.end();
}

Agent *World::getNextAgent()
{
	Agent *tmp = currentAgentIt->second;
	currentAgentIt++;
	return tmp;
}

bool World::hasRoadStart(unsigned roadId)
{
	return roadStarts.find(roadId) != roadStarts.end();
}

RoadStart *World::getRoadStart(unsigned roadId)
{
	if(hasRoadStart(roadId))
		return roadStarts[roadId];
	else
	{
		Communicator::getLogFile() << "Error in World::getRoadStart(unsigned roadId): "
		                           << "Requested RoadStart with id " 
		                           << roadId << " which does not exist" << endl;
		throw "World::getRoadStart(unsigned roadId): Requested inexistant RoadStart";
	}
}

bool World::hasRoadEnd(unsigned roadId)
{
	return roadEnds.find(roadId) != roadEnds.end();
}

RoadEnd *World::getRoadEnd(unsigned roadId)
{
	if(hasRoadEnd(roadId))
		return roadEnds[roadId];
	else
	{
		Communicator::getLogFile() << "Error in World::getRoadEnd(unsigned roadId): "
		                           << "Requested RoadEnd with id " 
		                           << roadId << " which does not exist" << endl;
		throw "World::getRoadEnd(unsigned roadId): Requested inexistant RoadEnd";
	}
}

unsigned World::countAgents()
{
	return agents.size();
}

unsigned World::countRoadStarts()
{
	return roadStarts.size();
}

unsigned World::countRoadEnds()
{
	return roadEnds.size();
}

void World::directlyConnectInternalLinks()
{
	unsigned numDirectlyConnected = 0;
	// iterate throug all road ends and directly connect them to road starts where possible
	for(map<unsigned int, RoadEnd*>::iterator rei = roadEnds.begin();
	    rei != roadEnds.end(); ++rei)
	{
		RoadEnd *currentRoadEnd = (*rei).second;
		if(currentRoadEnd->isDirectlyConnectable())
		{
			currentRoadEnd->directlyConnectToRoadStart();
			if(currentRoadEnd->isDirectlyConnected())
				++numDirectlyConnected;
			else
			{
				Communicator::getLogFile() << "failed to make a direct connection for road end "
				                           << currentRoadEnd->getId() << endl;
				throw;
			}
		}
	}
	if(DEBUG)
		Communicator::getLogFile() << "created a direct connection for " 
		                           << numDirectlyConnected << " roads." << endl;
}


void World::startAgents()
{
	for(map<unsigned int, Agent*>::iterator ai = agents.begin();
	    ai != agents.end(); ++ai)
	{
		(*ai).second->startActing();
	}
}

void World::startRoads()
{
	for(map<unsigned int, RoadStart*>::iterator rsi = roadStarts.begin();
	    rsi != roadStarts.end(); ++rsi)
		(*rsi).second->registerInitialTimerAtClock();
	for(map<unsigned int, RoadEnd*>::iterator rei = roadEnds.begin();
	    rei != roadEnds.end(); ++rei)
		(*rei).second->registerInitialTimerAtClock();
	
}

double World::getFirstSyncTime()
{
	return firstSyncTime;
}

void World::noticeAgentFinished(unsigned agentId)
{
	++numFinishedAgents;
}

unsigned World::getNumFinishedAgents() const
{
	return numFinishedAgents;
}



}
}
