/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "toolsglobal.h"
#include "link.h"
#include "node.h"
#include <cmath>
#include <iostream>

namespace allsim 
{
namespace tools
{

///\class Link
///\brief A class that represents a link in a general network.
///
/// A Link stores its corresponding to and from nodes and the attributes
/// found in network files
///
///\author David Charypar

/// Standard constructor used during network read in
Link::Link(unsigned id, Node *fromNode, Node *toNode, 
           double length, double flowCapacity, double freeSpeed, unsigned numLanes)
:id(id),
fromNode(fromNode), toNode(toNode),
length(length),
flowCapacity(flowCapacity),
freeSpeed(freeSpeed),
numLanes(numLanes)
{
	if(flowCapacity <= 0.0)
	{
		cerr << "id: " << id << endl;
		cerr << "length: " << length << endl;
		cerr << "flowCapacity: " << flowCapacity << endl;
		cerr << "freeSpeed: " << freeSpeed << endl;
		cerr << "numLanes: " << numLanes << endl;
		abort();
		throw "Error in Link::Link(): flow capacity has invalid value";
	}
}

/// Destructor.
Link::~Link()
{
}

/// Sets a new fromNode
void Link::setFromNode(Node *newFromNode)
{
	fromNode = newFromNode;
}

/// Sets a new toNode
void Link::setToNode(Node *newToNode)
{
	toNode = newToNode;
}

/// Returns the id of this link
unsigned int Link::getId() const
{
	return id;
}

/// Returns the fromNode of this link
Node *Link::getFromNode() const
{
	return fromNode;
}

/// Returns the toNode of this link 
Node *Link::getToNode() const
{
	return toNode;
}

/// Returns the length of this link (in meters)
double Link::getLength() const
{
	return length;
}

/// returns the flow capacity of this link (in vehicles per second)
double Link::getFlowCapacity() const
{
	return flowCapacity;
}

/// Returns the physical spaces on the link (in meters)
///
/// The space on the link is computed as number of lanes times the length of the link (in meters)
double Link::getSpace() const
{
	return numLanes * length;
}

/// Returns the free speed on the link (in meters per second)
double Link::getFreeSpeed() const
{
	return freeSpeed;
}

/// Returns the number of lanes of the link
int Link::getNumLanes() const
{
	return numLanes;
}

/// Returns the free speed travel time on the link (in seconds)
///
/// The free speed travel time is computed as lenth of link divided by free speed.
/// The resulting time is measured in seconds.
double Link::getFreeSpeedTravelTime() const
{
	return length/freeSpeed;
}

/// Returns the minimal time gap betwen two cars directly following each other
///
/// The minimum gap is computed as 1 / flow capacity and measured in seconds.
double Link::getMinimalGap() const
{
	return 1.0 / flowCapacity;
}

}
}
