/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "network.h"
#include "node.h"
#include "link.h"
#include <iostream>
#include <sstream>

namespace allsim 
{
namespace tools
{
/// \class Network
/// \author David Charypar
/// \brief A network class that stores the links and nodes of a network.
///
/// The Node objects and Link objects are stored and can be retrieved using the id of the objects.
/// Usually an empty network is constructed and the nodes and link are added using addNode() and addLink().
/// The destructor and delete() delete all the links and nodes stored in the datastructure.


/// Default constructor
Network::Network()
{
}

/// Copy constructor performing a deep copy.
///
/// All nodes and links present in \b source are copied and connected accordingly. 
/// That means that the copy constructor creates an independent object includeing all nodes
/// and links. The source can be safely destroyed afterwards 
Network::Network(Network const &source)
{
	// create a copy of each node
	for(map<unsigned, Node*>::const_iterator ni = source.nodes.begin(); 
	    ni != source.nodes.end(); ++ni)
	{
		Node *newNode = new Node(*ni->second);
		nodes[newNode->getId()] = newNode;
	}
	// create a copy of each link including connections to nodes
	for(map<unsigned, Link*>::const_iterator li = source.links.begin(); 
	    li != source.links.end(); ++li)
	{
		Link *csl = li->second;
		Node *newFromNode = nodes[csl->getFromNode()->getId()];
		Node *newToNode = nodes[csl->getToNode()->getId()];
		Link *newLink = new Link(csl->getId(), newFromNode, newToNode,
		                         csl->getLength(), csl->getFlowCapacity(),
		                         csl->getFreeSpeed(), csl->getNumLanes());
		links[newLink->getId()] = newLink; 
		// register link with from and to nodes.
		newFromNode->addOutgoingLink(newLink);
		newToNode->addIncomingLink(newLink);
	}
}

/// Destructor deletes all nodes and links present in the network.
Network::~Network()
{
	clear();
}

/// Deletes all nodes and links present in the network.
///
/// The node and link objects are physically deleted.
void Network::clear()
{
	// the network is responsible for destroying the links nodes 
	// that were created by the network reader.
	// --> loop over the links and nodes maps
	for(map<unsigned, Link*>::const_iterator li = links.begin(); 
	    li != links.end(); ++li)
	{
		delete li->second;
	}
	links.clear();

	for(map<unsigned, Node*>::const_iterator ni = nodes.begin(); 
	    ni != nodes.end(); ++ni)
	{
		delete ni->second;
	}
	nodes.clear();
}

/// Adds a node to the network.
void Network::addNode(Node *node)
{
	nodes[node->getId()] = node;
}

/// Adds a link to the network.
void Network::addLink(Link *link)
{
	links[link->getId()] = link;
}

/// Returns the Node with id \b nodeId.
///
/// If no node with this id exists an exception is thrown.
Node *Network::getNodeWithId(unsigned nodeId)
{
	if(nodes.find(nodeId) == nodes.end())
	{
		cerr << "hallo" << endl;
		stringstream tmp;
		tmp << "Node with node id " << nodeId << " does not exist in network";
		cerr << tmp.str() << endl;
		throw(tmp.str().c_str());
		return NULL;
	}
	else
		return nodes[nodeId];
}

/// Returns the link with id \b linkId.
///
/// if no link with this id exists an exception is thrown.
Link *Network::getLinkWithId(unsigned linkId)
{
	if(links.find(linkId) == links.end())
		return NULL;
	else
		return links[linkId];
}

/// Returns the total number of nodes in the network.
unsigned Network::getNumNodes() const
{
	return nodes.size();
}

/// Returns the total number of links in the network.
unsigned Network::getNumLinks() const
{
	return links.size();
}

/// Computes and returns the total space available in the network (in meters).
double Network::getSpaceOnRoads()
{
	double sum = 0.0;
	for(map<unsigned, Link*>::const_iterator li = links.begin(); 
	    li != links.end(); ++li)
	{
		sum += (li->second)->getSpace();
	}
	return sum;
}

/// Returns a vector containing all links present in the network.
vector<Link*> Network::getLinksVector()
{
	vector<Link*> allLinks;

	for(map<unsigned, Link*>::const_iterator li = links.begin(); 
	    li != links.end(); ++li)
	{
		allLinks.push_back(li->second);
	}

	return allLinks;
	
}
}
} // end namespace allsim
