/***************************************************************************
 *   Copyright (C) 2005-2007 by David Charypar   *
 *   david.charypar@switzerland.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "networkreader.h"
#include "network.h"
#include "node.h"
#include "link.h"
#include <cassert>
#include <expat.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <cstring>

using namespace std;


namespace allsim 
{
namespace tools
{

// static member variables...
unsigned NetworkReader::depth;
char NetworkReader::Buff[BUFFSIZE];
Network *NetworkReader::currentNetwork(NULL);
double NetworkReader::capPeriod(3600.0);



Network *NetworkReader::loadNetworkFromXML(char const *filename)
{
	NetworkReader::currentNetwork = new Network();
	XML_Parser p = XML_ParserCreate(NULL);
	if (! p) {
		cerr << "Couldn't allocate memory for parser" << endl;
		exit(-1);
	}

	XML_SetElementHandler(p, NetworkReader::startHandler, NetworkReader::endHandler);
	XML_SetCharacterDataHandler(p, NetworkReader::characterDataHandler);
	ifstream xmlStream(filename);
	

	bool done = false;
	do 
	{
		int len;
		xmlStream.getline(Buff, BUFFSIZE);
		len = xmlStream.gcount();
		done = !xmlStream.good();

		if (! XML_Parse(p, Buff, max(len-1,0), done)) 
		{
			cerr << "Parse error at line " << XML_GetCurrentLineNumber(p) << ":" << endl;
			cerr << XML_ErrorString(XML_GetErrorCode(p))<< endl;
			abort();
		}
  	}
	while(!done);

//	cerr << "read in " << currentNetwork->getNumNodes() << " nodes and ";
//	cerr << currentNetwork->getNumLinks() << " links." << endl;

	// freeing the XML-Parser
	XML_ParserFree(p);

	return currentNetwork;	
}

void NetworkReader::startHandler(void *, const char *el, const char **attr) 
{
	// decide based on the tag string which handler method to call
	if(strcmp(el, "network")==0)
		networkHandlerStart(attr);
	else
		if(strcmp(el, "node")==0)
			nodeHandlerStart(attr);
	else
		if(strcmp(el, "links")==0)
			linksHandlerStart(attr);
	else
		if(strcmp(el, "link")==0)
			linkHandlerStart(attr);
	depth++;
}

void NetworkReader::endHandler(void *, const char *el) 
{
	depth--;
	
	// decide based on the tag string which handler method to call
	if(strcmp(el, "network")==0)
		networkHandlerEnd();
	else
	if(strcmp(el, "node")==0)
		nodeHandlerEnd();
	else
	if(strcmp(el, "links")==0)
		linksHandlerEnd();
	else
	if(strcmp(el, "link")==0)
		linkHandlerEnd();
}

void NetworkReader::characterDataHandler(void *, const char *s, int len)
{
	if(depth < 3) return;

	char *myBuf = new char[len+1];
	strncpy(myBuf, s, len);
	myBuf[len] = '\0';
	for (unsigned i = 0; i < depth; i++)
		cout << "  ";
	cerr << myBuf <<  endl;
	delete myBuf;
}

void NetworkReader::networkHandlerStart(const char **)
{
	// the network was freshly created. Therefore, nothing has to be done here.
}

void NetworkReader::networkHandlerEnd()
{
	// nothing to be done to finnish the creation of the network
}


void NetworkReader::nodeHandlerStart(const char **attr)
{
	unsigned int nodeId = 0;
	double nodeXCoord, nodeYCoord;
	// store all attributes of the node
	for (int i = 0; attr[i]; i += 2)
	{
		if(strcmp(attr[i], "id")==0)
			nodeId = atoi(attr[i+1]);
		else
			if(strcmp(attr[i], "x")==0)
				nodeXCoord = atof(attr[i+1]);
			else
				if(strcmp(attr[i], "y")==0)
					nodeYCoord = atof(attr[i+1]);
	}
	// create the node and store it in the network
	Node *newNode = new Node(nodeId, nodeXCoord, nodeYCoord);
	currentNetwork->addNode(newNode);
}

void NetworkReader::nodeHandlerEnd()
{
	// nothing to be done here right now
}

void NetworkReader::linksHandlerStart(const char **attr)
{
	bool hasCapPeriod = false;
	// store capacity period
	for (int i = 0; attr[i]; i += 2)
	{
		if(strcmp(attr[i], "capperiod")==0)
		{
			capPeriod = convertTimeStringToTime(attr[i+1]);
			hasCapPeriod = true;
		}
	}
//	cerr << "using capperiod "  <<  capPeriod << endl;
	assert(hasCapPeriod);
}

void NetworkReader::linksHandlerEnd()
{
}


void NetworkReader::linkHandlerStart(const char **attr)
{
	unsigned int linkId = 0, linkFromNode = 0, linkToNode = 0, linkNumLanes = 0;
	double linkLength, linkCapacity, linkFreeSpeed;
	// store all attributes of the link
	for (int i = 0; attr[i]; i += 2)
	{
		if(strcmp(attr[i], "id")==0)
			linkId = atoi(attr[i+1]);
		else
		if(strcmp(attr[i], "from")==0)
			linkFromNode = atoi(attr[i+1]);
		else
		if(strcmp(attr[i], "to")==0)
			linkToNode = atoi(attr[i+1]);
			else
		if(strcmp(attr[i], "length")==0)
			linkLength = atof(attr[i+1]);
		else
		if(strcmp(attr[i], "capacity")==0)
			linkCapacity = atof(attr[i+1]);
		else
		if(strcmp(attr[i], "freespeed")==0)
			linkFreeSpeed = atof(attr[i+1]);
		else
		if(strcmp(attr[i], "permlanes")==0)
			linkNumLanes = atoi(attr[i+1]);
	}
	// get the pointer to the from and to nodes from the network:
	Node *fromNode = currentNetwork->getNodeWithId(linkFromNode);
	Node *toNode = currentNetwork->getNodeWithId(linkToNode);

	// create the link and store it in the network
	
	Link *newLink = new Link(linkId, fromNode, toNode, linkLength, 
	                         linkCapacity / capPeriod,
	                         linkFreeSpeed, linkNumLanes);

	// register this link as outgoing at the fromNode
	fromNode->addOutgoingLink(newLink);
	// register this link as incoming at the toNode
	toNode->addIncomingLink(newLink);
	
	// register this link with the network
	currentNetwork->addLink(newLink);
}

void NetworkReader::linkHandlerEnd()
{
	// nothing to be done here right now
}


}
}
